"""abed -- ABundances file EDitor window."""

__all__ = ["XFileAbonds"]

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from . import a_WFileAbonds
import os
from ._shared import *
import a99
import pyfant
import f311

################################################################################
class XFileAbonds(QMainWindow):
    """
    Args:
      parent=None: nevermind
      file_abonds (optional)-- FileAbonds instance
    """

    def __init__(self, parent=None, file_abonds=None):
        ## State variables
        QMainWindow.__init__(self, parent)
        self.flag_changed = False
        self.save_dir = "."
        me = self.editor = a_WFileAbonds.WFileAbonds()
        me.setFont(a99.MONO_FONT)
        me.changed.connect(self.on_edited)
        me.setFocus()
        # self.setWindowTitle(title)
        self.setCentralWidget(me)


        # # Status bar
        self.labelStatus = QLabel("")
        sb = self.statusBar()
        sb.insertWidget(0, self.labelStatus, 0)


        # # Menu bar
        b = self.menuBar()
        m = self.menu_file = b.addMenu("&File")
        self.act_save = ac = m.addAction("&Save")
        ac.setShortcut("Ctrl+S")
        ac.triggered.connect(self.on_save)
        self.act_save_as = ac = m.addAction("Save &as...")
        ac.setShortcut("Ctrl+Shift+S")
        ac.triggered.connect(self.on_save_as)
        self.act_export_dissoc = ac = m.addAction("Export &dissoc file...")
        ac.setStatusTip("Saves dissoc.dat file")
        ac.setShortcut("Ctrl+Shift+D")
        ac.triggered.connect(self.on_export_dissoc)
        self.act_export_turbospectrum = ac = m.addAction("Generate text for &TurboSpectrum")
        ac.setStatusTip("Opens window containing atomic numbers and abundances to be copy-pasted")
        ac.setShortcut("Ctrl+Shift+T")
        ac.triggered.connect(self.on_export_turbospectrum)
        m.addSeparator()
        ac = m.addAction("&Quit")
        ac.setShortcut("Ctrl+Q")
        ac.triggered.connect(self.close)

        rect = QApplication.desktop().screenGeometry()
        self.setGeometry(0, 0, 400, rect.height())
        a99.place_left_top(self)

        if file_abonds is not None:
            self.load(file_abonds)

    def load(self, x):
        assert isinstance(x, pyfant.FileAbonds)
        self.editor.load(x)
        self.update_window_title()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Override

    def closeEvent(self, evt):
        a99.are_you_sure(self.flag_changed, evt, self)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots

    def on_save(self, _):
        self.disable_save_actions()
        try:
            if not self.editor.flag_valid:
                a99.show_error(PARAMS_INVALID)
            else:
                self.save()
        except Exception as e:
            a99.show_error(str(e))
            raise
        finally:
            self.enable_save_actions()

    def on_save_as(self, _):
        self.disable_save_actions()
        try:
            if self.editor.f:
                if not self.editor.flag_valid:
                    a99.show_error(PARAMS_INVALID)
                else:
                    new_filename = QFileDialog.getSaveFileName(self, "Save file",
                     self.save_dir, "*.dat")
                    if new_filename:
                        self.save_dir, _ = os.path.split(str(new_filename))
                        self.save_as(new_filename)
        except Exception as e:
            a99.show_error(str(e))
            raise
        finally:
            self.enable_save_actions()

    def on_export_dissoc(self, _):
        self.disable_save_actions()
        try:
            if self.editor.f:
                if not self.editor.flag_valid:
                    a99.show_error(PARAMS_INVALID)
                else:
                    new_filename = QFileDialog.getSaveFileName(self, "Save file",
                     os.path.join(".", pyfant.FileDissoc.default_filename), "*.dat")[0]
                    if new_filename:
                        f = self.editor.f.get_file_dissoc()
                        f.title = "Created using abed.py"
                        f.save_as(new_filename)
        except Exception as e:
            a99.show_error(str(e))
            raise
        finally:
            self.enable_save_actions()


    def on_export_turbospectrum(self, _):
        w = f311.XText(self, self.editor.f.get_turbospectrum_str(), "Atomic number & abundance")
        w.show()

    def on_edited(self):
        self.flag_changed = True
        self.update_window_title()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Internals

    def enable_save_actions(self):
        self.act_save.setEnabled(True)
        self.act_save_as.setEnabled(True)

    def disable_save_actions(self):
        self.act_save.setEnabled(False)
        self.act_save_as.setEnabled(False)

    def save(self):
        if self.editor.f:
            self.editor.f.save_as()
            self.flag_changed = False
            self.update_window_title()

    def save_as(self, filename):
        if self.editor.f:
            self.editor.f.save_as(filename)
            self.flag_changed = False
            self.update_window_title()

    def update_window_title(self):
        self.setWindowTitle("abed -- %s%s%s" % (self.editor.f.filename,
          "" if not self.flag_changed else " (changed)",
          "" if self.editor.flag_valid else " (*invalid*)"))



from ..molconsts import *
from ..basic import state_to_str, SS_PLAIN
from f311 import DataFile
from collections import OrderedDict
import re
import a99
import io

__all__ = ["FileTRAPRBOutput", "FileTRAPRBInput", "TRAPRBInputState"]


class TRAPRBInputState(object):
    """Describes one of the two states in FileTRAPRBInput"""

    def __init__(self, title="", ni=0, ns=1, igraph=0, ienerg=0, istate=2, zmu=0., rmin=0.65, rmax=3.5,
                 delr=0.005, maxv=12, be=0., de=0., kdmaxv=12, neig=0, ev=None, bv=None):

        if ev is None:
            ev = []
        if bv is None:
            bv = []

        self.title = title

        # First 5 values we always leave as (0, 1, 0, 0, 2) [AAA]
        # = 0 ALLOWS DENSITY GRAPHS TO BE PRODUCED.OTHER VALUES DO NO
        self.ni = ni
        # GIVES ORTHOGONALITY TEST IF = 1
        self.ns = ns
        self.igraph = igraph
        self.ienerg = ienerg
        # = 2 FOR A TRANSITION = 1 FOR A SINGLE STATE
        self.istate = istate
        # "reduced mass" (m1*m2)/(m1+m2) (m*: atomic masses in the diatomic molecule) [AAA]
        self.zmu = zmu
        # MINIMUM RADIUS FOR CALCULATION OF WAVEFUNCTIONS
        self.rmin = rmin
        # MAXIMUM RADIUS FOR CALCULATION OF WAVEFUNCTIONS
        self.rmax = rmax
        # RADIAL INTERVAL FOR CALCULATION OF WAVEFUNCTIONS
        self.delr = delr
        # MAXIMUM QUANTUM NUMBER [i.e., vibrational, "v"] FOR CALCULATION OF RESULTS
        self.maxv = maxv
        # EQUILBRIUM ROTATIONAL CONSTANT [B_e from NIST or other source]
        self.be = be
        # DISSOCIATION ENERGY OR ESTIMATE THEREOF
        self.de = de
        # MAXIMUM QUANTUM NUMBER FOR CALCULATION OF KLEIN DUNHAM
        # POTENTIAL.MUST BE AT LEAST 2 AND NOT MORE THAN 30
        self.kdmaxv = kdmaxv
        # USED FOR WORK IN THE CONTINUUM, NUMBER OF ENERGY LEVELS
        # TO CALCULATE CONTINUUM WAVEFUNCTIONS AT
        self.neig = neig
        self.ev = ev
        self.bv = bv

    def __repr__(self):
        attrs = ["title", "ni", "ns", "igraph", "ienerg", "istate", "zmu", "rmin", "rmax", "delr", "maxv",
                 "be", "de", "kdmaxv", "neig", "ev", "bv"]
        return a99.make_code_readable("{}({})".format(self.__class__.__name__, ", ".join(["{}={}".format(x, repr(getattr(self, x))) for x in attrs])))


class FileTRAPRBInput(DataFile):
    """
    Input file for the TRAPRB Fortran code (which calculates Franck-Condon factors)

    Here is one annotated sample file:

    ```
                                                               ISTATE
                                                            IENERG  |
                                                         IGRAPH  |  |
                                                       NI NS  |  |  |
    TITLE---------------------------------------------  |  |  |  |  |
    1  OH A DOUBLET SIGMA UPPER STATE                   0  1  0  0  2

    ZMU-------RMIN------RMAX------DELR------
    0.9480871     0.750     3.500     0.005


    MAXV
    ---       BE        DE                                        KDMAXV NEIG
      |       ----------==========                                    ===---
     12           17.358  35000.00                                     12

    EV(32)
         00.00   1565.99   4554.82   7351.07   9959.53  12392.75  14671.06  16822.54
      18883.03  20896.17  22913.33  24993.67  27204.10  29619.32

    BV(32)
       17.3580   16.9606   16.1418   15.2910   14.4082   13.4934   12.5466   11.5678
       10.5570    9.5142    8.4394    7.3326    6.1938    5.0230

    FINAL STATE:
    SAME STRUCTURE AS BEFORE
    1  OH X DOUBLET PI  GROUND STATE                    0  1  0  0  2
    0.9480871      0.650   3.500     0.005
     12          18.9108  39162.18                                     12
          0.00   1847.73   5417.37   8821.40  12061.76  15139.53  18054.80  20806.55
      23392.49  25808.94  28050.69  30110.87  31980.82  33649.91


       18.9108   18.5504   17.8387   17.1366   16.4411   15.7493   15.0581   14.3645
       13.6655   12.9581   12.2394   11.5063   10.7557    9.9849
    ```

    References:

    [AAA] Talks with Prof. Amaury Augusto de Almeida

    [Jarmain&McCallum1970] Jarmain, W. R., and J. C. McCallum.
    "TRAPRB: a computer program for molecular transitions." University of Western Ontario (1970)
    """

    attrs = []

    def __init__(self):
        DataFile.__init__(self)

        self.states = [TRAPRBInputState(), TRAPRBInputState()]

    def from_molconsts(self, mc, maxv=None, ni=None, ns=None, igraph=None, ienerg=None, istate=None,
                       rmin=None, rmax=None, delr=None):
        """
        Calculates everything almost entirely based on a MolConsts object

        Information not present in mc may be passed as argument or let fall to same defaults as
        TRAPRBInputState.

        Args:
            mc: MolConsts instance
            maxv:
            ni:
            ns:
            igraph:
            ienerg:
            istate:
            rmin:
            rmax:
            delr:

        **Note**:  default to
                  TRAPRBInputState defaults

        **Note**: kdmaxv is made = maxv
        """
        assert isinstance(mc, MolConsts)

        def get_state_obj(omega_e, omega_ex_e, alpha_e, be, label, mult, spdf):
            # formulas given by [AAA]
            de = omega_e ** 2 / 4 / omega_ex_e
            ev = [0.] + [omega_e * (v + .5) - omega_ex_e * (v + .5) ** 2 for v in range(maxv + 1)]
            bv = [be] + [be - alpha_e * (v + .5) for v in range(maxv + 1)]

            st = TRAPRBInputState(
                title="{} {}".format(formula, state_to_str(label, mult, spdf, style=SS_PLAIN)),
                ni=ni, ns=ns, igraph=igraph, ienerg=ienerg, istate=istate,
                zmu=zmu,
                rmin=rmin, rmax=rmax, delr=delr,
                maxv=maxv,
                be=be,
                de=de,
                kdmaxv=kdmaxv,
                neig=neig,
                ev=ev,
                bv=bv,
            )
            return st

        default = TRAPRBInputState()

        maxv = maxv if maxv is not None else default.maxv
        ni = ni if ni is not None else default.ni
        ns = ns if ns is not None else default.ns
        igraph = igraph if igraph is not None else default.igraph
        ienerg = ienerg if ienerg is not None else default.ienerg
        istate = istate if istate is not None else default.istate
        rmin = rmin if rmin is not None else default.rmin
        rmax = rmax if rmax is not None else default.rmax
        delr = delr if delr is not None else default.delr

        zmu = mc["am"]*mc["bm"]/(mc["am"]+mc["bm"])

        kdmaxv = maxv

        # "No continuum to be studied". To understand this, look for this quoted string in FCF486.FOR
        neig = 0

        formula = mc["formula"]


        st0 = get_state_obj(
            mc["statel_omega_e"],
            mc["statel_omega_ex_e"],
            mc["statel_alpha_e"],
            mc["statel_B_e"],
            mc["from_label"],
            mc["from_mult"],
            mc["from_spdf"],
        )

        st1 = get_state_obj(
            mc["state2l_omega_e"],
            mc["state2l_omega_ex_e"],
            mc["state2l_alpha_e"],
            mc["state2l_B_e"],
            mc["to_label"],
            mc["to_mult"],
            mc["to_spdf"],
        )

        self.states = [st0, st1]

    def dumps(self):
        """Return string containing the same as file contents"""

        c = io.StringIO()
        self._write_to_stream(c)
        c.seek(0)
        return c.read()

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            self._write_to_stream(h)

    def _write_to_stream(self, h):
        self._write_state(h, self.states[0])
        self._write_state(h, self.states[1])

    def _write_state(self, h, st):
        h.write("{:50}{:3}{:3}{:3}{:3}{:3}\n".format(st.title, st.ni, st.ns, st.igraph, st.ienerg, st.istate))
        h.write("{:10g}{:10g}{:10g}{:10.3}\n".format(st.zmu, st.rmin, st.rmax, st.delr))
        h.write("{:3}{:7}{:10g}{:10g}{:36}{:3}\n".format(st.maxv, "", st.be, st.de, "", st.kdmaxv, st.neig))
        for i in range(4):
            h.write("".join("{:10g}".format(x) for x in st.ev[i*8:(i+1)*8])+"\n")
        for i in range(4):
            h.write("".join("{:10g}".format(x) for x in st.bv[i*8:(i+1)*8])+"\n")

    def _do_load(self, filename):
        with open(filename, "r") as h:
            self._read_state(h, self.states[0])
            self._read_state(h, self.states[1])

    def _read_state(self, h, st):

        def read32(h):
            """Used to read ev[] and bv[]. 4*8 values expected, but maybe not all present"""
            _ret = []
            for _ in range(4):
                _ret.extend(a99.str_vector(h))

            ret = [tfloat(x) for x in _ret]
            return ret

        # # Each "block" of information read is preceded by a sample of what needs to be read

        # 1  OH A DOUBLET SIGMA UPPER STATE                   0  1  0  0  2
        line = h.readline()
        assert isinstance(st, TRAPRBInputState)
        st.title = line[0:50]
        st.ni = tint(line[50:53])
        st.ns = tint(line[53:56])
        st.igraph = tint(line[56:59])
        st.ienerg = tint(line[59:62])
        st.istate = tint(line[62:65])

        # 0.9480871     0.750     3.500     0.005
        #
        line = h.readline()
        st.zmu = tfloat(line[0:10])
        st.rmin = tfloat(line[10:20])
        st.rmax = tfloat(line[20:30])
        st.delr = tfloat(line[30:40])

        #  12           17.358  35000.00                                     12
        line = h.readline()
        st.maxv = tint(line[0:3])
        st.be = tfloat(line[10:20])
        st.de = tfloat(line[20:30])
        st.kdmaxv = tint(line[66:69])
        st.neig = tint(line[69:72])

        #      00.00   1565.99   4554.82   7351.07   9959.53  12392.75  14671.06  16822.54
        #   18883.03  20896.17  22913.33  24993.67  27204.10  29619.32
        #
        #
        st.ev = read32(h)
        #    17.3580   16.9606   16.1418   15.2910   14.4082   13.4934   12.5466   11.5678
        #    10.5570    9.5142    8.4394    7.3326    6.1938    5.0230
        #
        #
        st.bv = read32(h)


def _tolerant(string, type_):
    """Tolerant int or float: empty string falls back to zero"""
    try:
        return type_(string)
    except ValueError:
        if len(string.strip()) == 0:
            return type_(0)
        raise


def tint(string):
    return _tolerant(string, int)


def tfloat(string):
    return _tolerant(string, float)


class FileTRAPRBOutput(DataFile):
    """
    Output file for the TRAPRB Fortran code (which calculates Franck-Condon factors)

    Usage: attribute "fcfs" is a dictionary accessed by key (vl, v2l)

    **History:**

    This file is the output of the TRAPRB Fortran code published in 1970 [Jarmain&McCallum1970]

    In 2015, I modified this Fortran code to compile with gfortran on request from
    Prof. Amaury Augusto de Almeida,.

    Bruno left several output files in his directory ATMOS:/wrk4/bruno/Mole/Fc
    containing tabulated FCFs for several molecules. The initial reason why I created this class
    was to read those files.

    References:

    [Jarmain&McCallum1970] Jarmain, W. R., and J. C. McCallum.
    "TRAPRB: a computer program for molecular transitions." University of Western Ontario (1970)
    """

    attrs = ["fcfs"]

    @property
    def num_lines(self):
        return len(self)

    def __len__(self):
        return len(self.fcfs)

    def __getitem__(self, item):
        return self.fcfs[item]

    def __init__(self):
        DataFile.__init__(self)

        self.fcfs = OrderedDict()


    def _do_load(self, filename):
        # File part of interest looks like this:

        #                                 V"   V""  FRANCK CONDON FACTOR     R-CENTROID
        #
        #                                  0   0    .8809749E+00             .1159111E+01
        #                                  0   1    .6597226E-01             .1573085E+01
        #                                  0   2    .4967269E-01             .1309538E+01
        #                                  0   3    .4893512E-06             .5487973E+02
        #                                  0   4    .1967671E-02             .1092197E+01
        #                                  0   5    .5098846E-03             .7344212E+00
        #                                  0   6    .3264353E-03             .8154120E+00
        #                                  0   7    .1959470E-03             .8020893E+00
        #                                  0   8    .1221533E-03             .7933001E+00
        #                                  0   9    .7849922E-04             .7871405E+00
        #                                  0  10    .5130633E-04             .7806426E+00
        #
        #                                 V"   V""  FRANCK CONDON FACTOR     R-CENTROID
        #
        #                                  1   0    .7950073E-01             .9007177E+00
        #                                  1   1    .7301335E+00             .1176593E+01
        #                                  1   2    .5329854E-01             .1873632E+01
        #                                  1   3    .1172008E+00             .1293195E+01
        #                                  1   4    .9687749E-03            -.1106724E+01
        #                                  1   5    .8699398E-02             .1027710E+01
        #                                  1   6    .3409328E-02             .7787252E+00
        #                                  1   7    .2211213E-02             .8098937E+00
        #                                  1   8    .1418452E-02             .8016906E+00
        #                                  1   9    .9225228E-03             .7913423E+00
        #                                  1  10    .6110833E-03             .7833238E+00
        # 1
        #
        #                                 V"   V""  FRANCK CONDON FACTOR     R-CENTROID
        #
        #                                  2   0    .1939492E-01             .1045730E+01
        #                                  2   1    .4528844E-01             .1349549E+01
        #                                  2   2    .1031371E+00             .1003365E+01
        #                                  2   3    .4823686E+00             .1247711E+01
        #                                  2   4    .7587573E-01             .2082369E+01
        #                                  2   5    .2123286E+00             .1379030E+01
        #                                  2   6    .2887595E-02            -.1045935E+01
        #                                  2   7    .1985530E-01             .9975332E+00
        #                                  2   8    .1103128E-01             .8153041E+00
        #                                  2   9    .7418684E-02             .8001242E+00
        #                                  2  10    .5176848E-02             .7976580E+00
        #

        # magic characters
        MAGIC = 'V"   V""  FRANCK CONDON FACTOR     R-CENTROID'

        # possible states for parsing file
        EXP_MAGIC = 0 # Expecting magic characters
        EXP_BLANK = 1 # Expecting blank line
        EXP_DATA = 2 # Expecting data, "1*" or blank line

        # regular expression for extraction the data
        rec = re.compile("\s*(\d+)\s*(\d+)\s*([\deE\.\+\-]+)")


        with open(filename, "r") as h:
            found = False
            state = EXP_MAGIC
            i = 0
            try:
                for line in h:
                    i += 1
                    if state == EXP_MAGIC:
                       if line.strip() == MAGIC:
                           state = EXP_BLANK
                           found = True

                    elif state == EXP_BLANK:
                        if line.strip() == "":
                            state = EXP_DATA
                    elif state == EXP_DATA:
                        if line.strip() == "" or line[0] == "1":
                            state = EXP_MAGIC
                        else:
                            m = rec.match(line)
                            if m is None:
                                raise RuntimeError("Could not extract (vl, v2l, fcf)")
                            vl, v2l, fcf = m.groups()
                            vl = int(vl)
                            v2l = int(v2l)

                            if (vl, v2l) in self.fcfs:
                                a99.get_python_logger().warning("Repeated (vl, v2l): ({}, {})".format(vl, v2l))

                            fcf = float(fcf)
                            self.fcfs[(vl, v2l)] = fcf
            except Exception as e:
                raise RuntimeError("Error in line {}".format(i)) from e


            if not found:
                raise RuntimeError("File does not appear to be a FileTRAPRBOutput")




"""Widget to edit a series of points on top of a spectrum"""

from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *
import a99
import matplotlib.pyplot as plt
import f311


__all__ = ["WContinuum"]

_COLOR_LINE = (.91, .13, 0)
_COLOR_MARKER = (.92, .132, 0)
_COLOR_SPECTRUM = (0., .2, .9)
_LINESTYLE_CONTINUUM = "solid"
_LINESTYLE_SPECTRUM = "dashed"
_LINEWIDTH = 1
_MARKER = "s"


class WContinuum(a99.WBase):

    @property
    def points(self):
        return self._points

    @points.setter
    def points(self, value):
        self._points = value
        self._update_self()
        self._update_gui()

    @property
    def spectrum(self):
        return self._spectrum

    @spectrum.setter
    def spectrum(self, value):
        self._spectrum = value
        self._update_self()
        self._update_gui()

    def __init__(self, parent_form, spectrum=None, points=None):
        a99.WBase.__init__(self, parent_form)

        self._spectrum = spectrum
        self._points = points

        lw0 = self.layout_main = QVBoxLayout()
        self.setLayout(lw0)

        # Toolbar

        lwtb = self.keep_ref(QHBoxLayout())
        lw0.addLayout(lwtb)
        ###
        laa = self.label_llzero = QLabel("No tools yet")
        lwtb.addWidget(laa)
        ###
        lwtb.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # Matplotlib area

        wm = a99.keep_ref(QWidget())
        # a99.set_margin(wm, 0)
        lw0.addWidget(wm)
        self.figure, self.canvas, self.lfig = a99.get_matplotlib_layout(wm)
        self.canvas.mpl_connect('button_press_event', self._on_plot_click)

        self._update_self()
        self._update_gui()

    def _sth_changed(self):
        self._update_fobj()
        self.changed.emit()

    def _update_self(self):
        """Changes to internal state sensible to current internal state"""
        sp = self._spectrum
        if self._points is None and sp is not None:
            if False:
                y0, y1 = sp.y[0], sp.y[-1]
            else:
                y0, y1 = 1., 1.
            self._points = [(sp.l0, y0), (sp.lf, y1)]

    def _update_gui(self):
        """Changes to GUI sensible to internal state"""
        self._draw()

    def _draw(self):
        fig = self.figure
        fig.clear()
        plt.figure(fig.number)

        try:
            a99.format_BLB()

            # Spectrum
            sp = self._spectrum
            if sp is not None:
                setup = f311.PlotSpectrumSetup()
                plt.plot(sp.x, sp.y, label=str(sp.title), color=_COLOR_SPECTRUM,
                         linestyle=_LINESTYLE_SPECTRUM, linewidth=_LINEWIDTH)
                plt.xlabel(setup.fmt_xlabel.format(sp))

                xmin, xmax, ymin, ymax, xspan, yspan = f311.calc_max_min([sp])
                ymin, ymax = min(0., ymin), max(1., ymax)

                _T = 0.02  # fraction of extra space on left, right, top, bottom of graphics
                plt.xlim([xmin - xspan * _T, xmax + xspan * _T])
                plt.ylim([ymin - yspan * _T, ymax + yspan * _T])


            # Continuum
            if self._points is not None:
                xx, yy = zip(*self._points)
                plt.plot(xx, yy, c=_COLOR_LINE, linestyle=_LINESTYLE_CONTINUUM, linewidth=_LINEWIDTH)

                plt.plot(xx, yy, c=_COLOR_MARKER, linestyle="None", marker=_MARKER)

            try:
                fig.tight_layout()
            except:
                a99.get_python_logger().exception("WContinuum._draw()")

        except:
            a99.get_python_logger().exception("WContinuum._draw()")
        finally:
            self.canvas.draw()


    def _on_plot_click(self, event):
        print('%s click: button=%d, x=%d, y=%d, xdata=%f, ydata=%f' %
              ('double' if event.dblclick else 'single', event.button,
               event.x, event.y, event.xdata, event.ydata))

        x, y = event.xdata, event.ydata
        sp = self.spectrum
        print("XXXXXXXXX {} YYYYYYYYYY {}".format(x, y))

        if sp is not None:

            idx = a99.index_nearest(sp.x, x)
            print("Lambda is actually {}".format(sp.x[idx]))

        if event.button == 1:
            self._insert_point(x, y)
        elif event.button == 3:
            self._delete_nearest(x, y)

    def _insert_point(self, x, y):
        p = (x, y)
        pp = self._points
        xx, _ = zip(*pp)
        idx = a99.BSearchCeil(xx, x)
        if idx == -1:
            pp.append(p)
        else:
            pp.insert(idx, p)
        self._update_gui()

    def _delete_nearest(self, x, y):
        """Deletes point nearest to (x, y) (actually ignores the y)"""

        sp = self._spectrum
        if sp is not None:
            pp = self._points

            if len(pp) < 3:
                self.add_log_error("Need at least two points")
                return

            xx, _ = zip(*pp)
            index = a99.index_nearest(xx, x)
            del pp[index]

            self._update_gui()



__all__ = ["FileMain"]

import a99
from f311 import DataFile


NINE_HUNDRED = 900


@a99.froze_it
class FileMain(DataFile):
    """
    PFANT Main Stellar Configuration

    Imitates the logic of reader_main.f90::read_main().

    Attributes match reader_main.f90::main_* (minus the "main_" prefix)
    """

    default_filename = "main.dat"
    attrs = ["titrav", "pas", "fwhm", "ivtot", "vvt", "tolv",
             "teff", "glog", "asalog", "nhe", "ptdisk", "mu",
             "flprefix", "llzero", "llfin", "aint"]
    editors = ["mained.py", "x.py"]

    def __init__(self):
        DataFile.__init__(self)
        self.titrav = None
        self.ecrit = None
        self.pas = None
        self.echx = None
        self.echy = None
        self.fwhm = None
        self.ivtot = None
        self.vvt = None
        self.tolv = None
        self.teff = None
        self.glog = None
        self.asalog = None
        self.nhe = None
        self.inum = None
        self.ptdisk = None
        self.mu = None
        self.afstar = None
        self.xxcor = None
        self.flprefix = None
        self.llzero = None
        self.llfin = None
        self.aint = None
        self.filetohy = None

    def _do_load(self, filename):
        """Loads from file."""

        with open(filename, "r") as h:
            self.titrav = h.readline().strip()
            self.ecrit, self.pas, self.echx, self.echy, self.fwhm = a99.str_vector(h)

            vvt = float(h.readline())
            self.ivtot = 1
            if vvt <= NINE_HUNDRED:
                self.vvt = [vvt]
            else:
                _ivtot = int(h.readline())
                self.tolv = a99.float_vector(h)
                self.vvt = a99.float_vector(h)
                self.ivtot = len(self.vvt)

                if not (self.ivtot == len(self.tolv) == len(vvt)):
                    raise RuntimeError("ivtot must match len(tolv) must match len(vvt)")

            self.teff, self.glog, self.asalog, self.nhe, self.inum = a99.str_vector(h)
            self.ptdisk, self.mu = a99.str_vector(h)
            self.afstar = float(h.readline())
            self.xxcor = a99.float_vector(h)
            self.flprefix = h.readline().strip()
            self.llzero, self.llfin, self.aint = a99.float_vector(h)

            self.filetohy = []
            for s in h.readlines():
                s = s.strip()
                if len(s) > 0:
                    self.filetohy.append(s)

        # remaining conversions
        self.ecrit, self.ptdisk = list(map(a99.str2bool, (self.ecrit, self.ptdisk)))
        self.pas, self.echx, self.echy, self.fwhm, self.teff, self.glog, \
        self.asalog, self.nhe, self.mu = \
            list(map(float, (self.pas, self.echx, self.echy, self.fwhm, self.teff,
                             self.glog, self.asalog, self.nhe, self.mu)))
        self.inum = int(self.inum)

    def _do_save_as(self, filename):
        """Saves to file."""
        assert isinstance(self.vvt, list), "vvt must be list!"
        list2str = lambda l: " ".join([str(x) for x in l])
        with open(filename, "w") as h:
            a99.write_lf(h, "%s" % self.titrav)
            a99.write_lf(h, "%s %s %s %s %s" % (a99.bool2str(self.ecrit),
                                            self.pas, self.echx, self.echy, self.fwhm))
            a99.write_lf(h, "%s" % self.vvt[0])
            if self.vvt[0] > NINE_HUNDRED:
                a99.write_lf(h, "%s" % len(self.vvt))  # self.ivtot)
                a99.write_lf(h, list2str(self.tolv))
                a99.write_lf(h, list2str(self.vvt))

            a99.write_lf(h, list2str([self.teff, self.glog, self.asalog, self.nhe, self.inum]))
            a99.write_lf(h, list2str([a99.bool2str(self.ptdisk), self.mu]))
            a99.write_lf(h, str(self.afstar))
            a99.write_lf(h, list2str(self.xxcor))
            a99.write_lf(h, self.flprefix)
            a99.write_lf(h, list2str([self.llzero, self.llfin, self.aint]))
            for filetoh in self.filetohy:
                a99.write_lf(h, filetoh)



#!/usr/bin/env python

"""Main configuration file editor."""

import sys
import argparse
import logging
import a99
import pyfant


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='main configuration file name', default='main.dat', nargs='?')
    args = parser.parse_args()

    m = pyfant.FileMain()
    m.load(args.fn)
    app = a99.get_QApplication([])
    form = pyfant.XFileMain()
    form.load(m)
    form.show()
    sys.exit(app.exec_())



from .filemod import *
from .filedissoc import *
from .fileabonds import *
from .fileabsoru2 import *
from .fileabxfwhm import *
from .fileatoms import *
from .filehmap import *
from .filemain import *
from .filemolecules import *
from .fileoptions import *
from .filepartit import *
from .filetoh import *
from .filevald3 import *
from .filemoldb import *
from .filehitrandb import *
from .filekuruczmol import *
from .filepleztio import *
from .filemolconsts import *
from .fileconfconvmol import *
from .filetraprb import *
from .morespectra import *
from .fileplezlinelist import *
from .filebrooke2014 import *
from .filemollist import *
from .filehitran160 import *
from .filecojm1998 import *


#!/usr/bin/env python
"""
Runs pfant and nulbad in "multi mode" (equivalent to Tab 4 in "x.py")

This script runs spectral synthesis and convolutions with Gaussian profile for
several combinations of (atomic abundance, FWHM)
"""

import argparse
import pyfant
import a99
import logging


a99.logging_level = logging.INFO
a99.flag_log_file = True

DEFAULT_SESSION_ID = pyfant.MULTISESSION_PREFIX+"<i>"


if __name__ == "__main__":
    # Configuration for Python logging messages.
    logger = a99.get_python_logger()

    # Parser command-line arguments
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    names = pyfant.Conf().opt.get_names() # option names
    for name in names:
        # name = name.replace('_', '-')
        parser.add_argument("--"+name, type=str, help='')

    parser.add_argument("-f", "--fn_abxfwhm", type=str, default=pyfant.FileAbXFwhm.default_filename,
                        help="Name of file specifying different abundances and FWHM's")
    parser.add_argument("-s", "--custom_session_id", type=str, default=DEFAULT_SESSION_ID,
                        help="Name of directory where output files will be saved")
    args = parser.parse_args()

    # Makes FileOptions object
    oopt = pyfant.FileOptions()
    for name in names:
        x = args.__getattribute__(name)
        if x is not None:
            oopt.__setattr__(name, x)

    oconf = pyfant.Conf()  # creates a Conf object just to use its get_file_*() methods
    omain = oconf.get_file_main(oopt)
    oabonds = oconf.get_file_abonds(oopt)

    oabxfwhm = pyfant.FileAbXFwhm()
    oabxfwhm.load(args.fn_abxfwhm)

    r = pyfant.MultiRunnable(omain, oabonds, oopt, oabxfwhm)
    if args.custom_session_id != DEFAULT_SESSION_ID:
        custom_id = args.custom_session_id
        if pyfant.get_custom_multisession_dirname(custom_id) == custom_id:
            # Understands that session dirname prefix must be cleared
            r.sid.id_maker.session_prefix_singular = ""
        r.sid.id = custom_id

    r.run()

    logger.info("Session directory: {}".format(r.sid.dir))



__all__ = ["FilePartit"]


import a99
from f311 import DataFile


class FilePartit(DataFile):
    """
    PFANT Partition Function

    Reader/writer not implemented (will be implemented  when there is the need for so)
    """

    default_filename = "partit.dat"

    def __init__(self):
        DataFile.__init__(self)

    def _do_load(self, filename):
        raise NotImplementedError("This class is a stub ATM")

    def _do_save_as(self, filename):
        raise NotImplementedError("This class is a stub ATM")



__all__ = ["FileDissoc"]


import fortranformat as ff
import a99
from f311 import DataFile
from ..basic import adjust_atomic_symbol

@a99.froze_it
class FileDissoc(DataFile):
    """PFANT Stellar Dissociation Equilibrium Information"""
    default_filename = "dissoc.dat"

    attrs = ['nmetal', 'nimax', 'elems', 'nelemx', 'ip', 'ig0', 'ig1', 'cclog',
             'nmol', 'mol', 'c', 'mmax', 'nelem', 'natom', 'eps', 'switer']

    def __init__(self):
        DataFile.__init__(self)
        # atom-related
        self.nmetal = None
        self.nimax = None
        self.elems = None
        self.nelemx = None
        self.ip = None
        self.ig0 = None
        self.ig1 = None
        self.cclog = None

        # molecule-related
        self.nmol = None
        self.mol = None
        self.c = None
        self.mmax = None
        self.nelem = None
        self.natom = None
        self.eps = None
        self.switer = None

        # inovation
        self.title = None

    def __len__(self):
        """Returns length of "ele" attribute."""
        return len(self.ele)

    def _do_load(self, filename):
        """Clears internal lists and loads from file."""

        with open(filename, "r") as h:
            fr = ff.FortranRecordReader('(2i5, 2f10.5, 1x, a100)')
            self.nmetal, self.nimax, self.eps, self.switer, self.title = fr.read(h.readline())
            self.title = self.title.strip()

            # atoms part
            self.elems, self.nelemx, self.ip, self.ig0, self.ig1, self.cclog = [], [], [], [], [], []
            fr = ff.FortranRecordReader('(a2, 2x, i6, f10.3, 2i5, f10.5)')
            for i in range(self.nmetal):
                symbol_, nelemx, ip, ig0, ig1, cclog = fr.read(h.readline())
                if len(symbol_.strip()) == 0:
                    raise RuntimeError("Doesn't seem to be a FileDissoc")

                self.elems.append(adjust_atomic_symbol(symbol_))
                self.nelemx.append(nelemx)
                self.ip.append(ip)
                self.ig0.append(ig0)
                self.ig1.append(ig1)
                self.cclog.append(cclog)

            if len(self.elems) == 0:
                raise RuntimeError("Doesn't seem to be a FileDissoc")


            # molecules part (remaining lines)
            self.mol, self.c, self.mmax, self.nelem, self.natom = [], [], [], [], []
            fr = ff.FortranRecordReader('(a3, 5x, e11.5, 4e12.5, i1, 4(i2,i1))')
            for line in h:
                if not line.strip():
                    # empty line is considered end-of-file
                    break
                vars = fr.read(line)

                self.mol.append(vars[0])
                self.c.append(vars[1:6])
                mmax = vars[6]
                self.mmax.append(mmax)
                self.nelem.append(vars[7::2][:mmax])
                self.natom.append(vars[8::2][:mmax])
            self.nmol = len(self.mol)




    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            # h.writelines([' %-2s%6.2f\n' % (self.ele[i], self.abol[i])
            #               for i in xrange(len(self))])
            # h.writelines(['1\n', '1\n'])

            a99.write_lf(h, "%5d%5d%10.5f%10.5f %s" % (self.nmetal, self.nimax, self.eps, self.switer, self.title))

            # atoms part
            fr = ff.FortranRecordReader('(a2, 2x, i6, f10.3, 2i5, f10.5)')
            for elems, nelemx, ip, ig0, ig1, cclog in \
                    zip(self.elems, self.nelemx, self.ip, self.ig0, self.ig1, self.cclog):
                elems = elems.upper().strip()  # to follow convention: right-aligned upper case
                a99.write_lf(h, '%2s  %6d%10.3f%5d%5d%10.5f' % (elems, nelemx, ip, ig0, ig1, cclog))


            for mol, c, mmax, nelem, natom in \
                    zip(self.mol, self.c, self.mmax, self.nelem, self.natom):
                mol = mol.upper().strip()  # to follow convention: right-aligned upper case
                l = ["%3s     %11.5e%12.5e%12.5e%12.5e%12.5e%1d" % tuple([mol]+c+[mmax])]
                for nelemm, natomm in zip(nelem, natom):
                    l.append("%2d%1d" % (nelemm, natomm))
                s = "".join(l)
                a99.write_lf(h, s)

            a99.write_lf(h, "")
            a99.write_lf(h, "")  # two blank lines to signal end-of-file





"""Molecular lines converter from HITRAN to PFANT format."""

__all__ = ["ConvHITRAN160"]

import a99, airvacuumvald as avv, sys
from . import Conv

class ConvHITRAN160(Conv):
    """Converts HITRAN molecular lines data to PFANT "sets of lines" using Einstein's coefficients

    Args:
        isowant: (=1) isotopologue ID according to HITRAN or literature e.g. [1]. Li et al. 2015 didn't follow
                 HITRAN's isotopologues codes for CO

        flag_jl_from_branch: whether to determine Jl from (branch, J2l) using rule:
            branch is P: Jl = J2l-1
            branch is Q: Jl = J2l
            branch is R: Jl = J2l+1

    There is a working example conv_hitran_co.py in examples directory

    References:
        [1] Li, G., Gordon, I.E., Rothman, L.S., Tan, Y., Hu, S.M., Kassi, S., Campargue, A. and Medvedev, E.S., 2015.
        Rovibrational line lists for nine isotopologues of the CO molecule in the X1Σ+ ground electronic state.
        The Astrophysical Journal Supplement Series, 216(1), p.15.

        [2] Plez's conversor translate_molec_linelists_v16.1.f

        [3] Jorge Melendez's work on OH: extraehitran.f
    """

    def __init__(self, *args, isowant=1, flag_jl_from_branch=False, **kwargs):
        super().__init__(*args, **kwargs)
        if not (1 <= isowant <= 9):
            raise ValueError(f"Invalid 'isowant': {isowant} (must be between 1 and 9)")
        self.isowant = isowant
        self.flag_jl_from_branch = flag_jl_from_branch

    def _make_sols(self, f):
        import pyfant
        assert isinstance(f, pyfant.FileHITRAN160)

        HITRAN = 1
        LI2015 = 2
        if isinstance(f, pyfant.FileHITRANHITRAN):
            flavor = HITRAN
        elif isinstance(f, pyfant.FileHITRANLi2015):
            flavor = LI2015
        else:
            raise RuntimeError(f"Cannot handle class {f.__class__.__name__}")

        log = self.log
        sols = self.sols

        n = log.n = len(f)
        if n == 0:
            raise RuntimeError("Zero lines found")

        deltak = self.molconsts.get_deltak()
        S2l = self.molconsts.get_S2l()

        if self.flag_filter_labels and flavor == LI2015:
            raise RuntimeError(f"Cannot filter labels (flag_filter_labels==True), as labels (e.g. \"A\", \"X\") "
                               f"are not provided by FileLi2015")

        if not self.flag_jl_from_branch and flavor == HITRAN:
            raise RuntimeError(f"Cannot determine Jl another way for FileHITRANHITRAN: flag_jl_from_branch must be activated")

        if self.flag_filter_labels:
            ref_from_label = self.molconsts["from_label"]
            ref_to_label = self.molconsts["to_label"]

        for i, line in enumerate(f.lines):
            try:
                if line.iso != self.isowant:
                    log.skip_reasons[f"isotopologue {line.iso}"] += 1
                    continue

                if self.flag_filter_labels:
                    if line.from_label != ref_from_label or line.to_label != ref_to_label:
                        log.skip_reasons[f"Wrong system: {line.from_label}-{line.to_label}"] += 1
                        continue

                if flavor == HITRAN:
                    J2l = line.J2l

                    # The following code was based on Jorge Melendez's extraehitran.f
                    # HITRAN does not have Jl
                    # branch examples: two letters, e.g. "PP", "QP", "SR", "PP", "OP"
                    if line.branch == "PP":
                        Jl = J2l-1
                    elif line.branch == "QQ":
                        Jl = J2l
                    elif line.branch == "RR":
                        Jl = J2l+1
                    else:
                        log.skip_reasons[f"Unhandled branch: {line.branch}"] += 1
                        continue
                    
                elif flavor == LI2015:
                    # This conversion of gp, gpp into Jl, J2l was taken from Plez's conversor [2]
                    # and seems to be correct for the Li 2015 CO file (I checked (vl=3, v2l=0) against the CO from 1998
                    # and J2l matches)
                    J2l = (line.gpp-1.)/2.

                    if not self.flag_jl_from_branch:
                        Jl = (line.gp-1.)/2.
                    else:
                        if line.branch == "P":
                            Jl = J2l-1
                        elif line.branch == "Q":
                            Jl = J2l
                        elif line.branch == "R":
                            Jl = J2l+1
                        else:
                            log.skip_reasons[f"Unhandled branch: {line.branch}"] += 1
                            continue

                else:
                    raise NotImplementedError()

                lambda_ = avv.vacuum_to_air(1e8/line.nu)

                SJ = self.get_sj_einstein(line.A, Jl, J2l, S2l, deltak, line.nu)

                self.append_line2(line.vl, line.v2l, lambda_, SJ, J2l, line.branch)

            except Exception as e:
                reason = a99.str_exc(e)
                log.skip_reasons[reason] += 1
                msg = "#{}{} line: {}".format(i+1, a99.ordinal_suffix(i+1), reason)
                log.errors.append(msg)
                if not self.flag_quiet:
                    a99.get_python_logger().exception(msg)


        return sols, log



from subprocess import PIPE
import subprocess
from .runnables import Runnable
import a99
import os
import pyfant

__all__ = ["TRAPRB"]

# TODO dismember with superclass should I want to create another executable, perhaps/probably start new lineage in a99!
class TRAPRB(Runnable):
    """
    Class representing the TRAPRB code [Jarmain&McCallum1970]

    References:

    [Jarmain&McCallum1970] Jarmain, W. R., and J. C. McCallum.
    "TRAPRB: a computer program for molecular transitions." University of Western Ontario (1970)
    """

    def __init__(self):
        Runnable.__init__(self)
        self.exe_path = "traprb"
        self.fn_input = "data.dat"
        self.fn_output = None
        self.flag_log_console = True
        self.flag_log_file = True

        self.__popen = None


    def run(self):
        """Runs executable.

        Blocking routine. Only returns when executable finishes running.
        """
        assert not self._flag_running, "Already running"
        assert not self._flag_finished, "Already finished"
        a99.get_python_logger().debug("Running %s '%s'" % (self.__class__.__name__.lower(), self.name))
        self.__run()

    def __run(self):
        """Called both from run() and run_from_combo()."""

        fn_output = self._get_fn_output()

        # Deletes output file if it already exists so that we don't have TRAPRB failing because of this
        if os.path.isfile(fn_output):
            os.unlink(fn_output)

        # Logging
        if self.flag_log_file:
            log_path = "traprb.log"
            if self.flag_log_console:
                stdout_ = a99.LogTwo(log_path)
            else:
                stdout_ = open(log_path, "w")
        else:
            if self.flag_log_console:
                stdout_ = subprocess.STD_OUTPUT_HANDLE
            else:
                stdout_ = None

        try:
            self._flag_running = True

            self.__popen = subprocess.run([self.exe_path], stdout=PIPE,
                input=bytes("{}\n{}\n".format(self.fn_input, fn_output), "ascii"))

            if stdout_ is not None:
                stdout_.write(self.__popen.stdout.decode("ascii"))

            if self.__popen.returncode != 0:
                raise pyfant.FailedError("%s failed (returncode=%s)" % (self.__class__.__name__.lower(), self.__popen.returncode))

        except Exception as e:
            flag_re_raise = True
            if self.__popen:
                if isinstance(e, pyfant.FailedError) and self._flag_killed:
                    # dismisses error if explicitly killed
                    flag_re_raise = False
                elif isinstance(e, IOError) and self.__popen.returncode == 0:
                    # Sometimes a IOError is raised even if Fortran executes
                    # successfully, so the error is dismissed
                    a99.get_python_logger().warning("Harmless error in TRAPRB")
                    flag_re_raise = False

            if flag_re_raise:
                self._error_message = e.__class__.__name__+": "+str(e)
                self._flag_error = True
                raise
        finally:
            self._flag_finished = True
            self._flag_running = False
            if self.__popen is not None:
                self.__returncode = self.__popen.returncode

            if stdout_ is not None:
                stdout_.close()




    def _get_fn_output(self):
        fn_output = self.fn_input + ".out" if self.fn_output is None else self.fn_output
        return fn_output

    def load_result(self):
        """Stores output in self.result["output"]"""
        f = pyfant.FileTRAPRBOutput()
        f.load(self._get_fn_output())
        self.result["output"] = f



#!/usr/bin/env python
"""
Converts TurboSpectrum atomic lines file to PFANT atomic lines file.

Certain elements are skipped.

Usage examples:

    To skip only H and He:
    > turbospectrum-to-atoms --skip "H, He"
"""

import argparse
import logging
import numpy as np
import a99
import pyfant
import sys


a99.logging_level = logging.INFO
a99.flag_log_file = True


DEFOUT = "atoms-untuned-<fn_input>"

TO_REMOVE = "H, He, F, Ne, P, Ar, Cl, As, Br, Kr, Xe"

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    parser.add_argument('fn_input', type=str, help='input file name', nargs=1)
    parser.add_argument('fn_output', type=str, help='output file name', nargs="?",
     default=DEFOUT)
    parser.add_argument('--min_algf', type=float, nargs='?', default=-7,
     help='minimum algf (log gf)')
    parser.add_argument('--max_kiex', type=float, nargs='?', default=15,
     help='maximum kiex')
    parser.add_argument("-s", "--skip", type=str, nargs='?', default=TO_REMOVE,
     help='list of elements to skip (use quotes and separate elements by commas)')

    args = parser.parse_args()
    logger = a99.get_python_logger()

    fn_out = args.fn_output
    if fn_out == DEFOUT:
        fn_out = "atoms-untuned-"+args.fn_input[0]

    logger.info("Converting file...")
    with open(args.fn_input[0], 'r') as file_:
        file_atoms = pyfant.turbospectrum_to_atoms(file_)

    n0 = file_atoms.num_lines
    logger.info("Number of lines before filtering: %d" % n0)

    logger.info("    Removing all (algf < %g)..." % args.min_algf)
    file_atoms.filter(lambda line: line.algf >= args.min_algf)
    logger.info("    Number of lines removed: %d" % (n0-file_atoms.num_lines))

    n0 = file_atoms.num_lines
    logger.info("    Filtering all (kiex > %g)..." % args.max_kiex)
    file_atoms.filter(lambda line: line.kiex <= args.max_kiex)
    logger.info("    Number of lines removed: %d" % (n0-file_atoms.num_lines))


    # Skipping elements (actually removing them after conversion is done)
    skip = [x.strip() for x in args.skip.split(",")]
    for ele in skip:
        n0 = file_atoms.num_lines
        logger.info("    Removing '{}' lines...".format(ele))
        file_atoms.remove_element(ele)
        logger.info("    Number of '{}' lines removed: {}".format(ele, n0-file_atoms.num_lines))

    logger.info("Number of lines after filtering: %d" % file_atoms.num_lines)

    logger.info("Saving file...")
    file_atoms.save_as(fn_out)
    print("File %s was successfully created." % fn_out)
    zz = file_atoms.zinf
    if len(zz) > 0:
        if np.all(zz == zz[0]):
            logger.warning("All zinf's set to %g. Now they must be fine-tuned with tune-zinf.py." % zz[0])



# todo plot "X" instead of line
# rename molecule


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT # as NavigationToolbar2QT
import matplotlib.pyplot as plt
import numpy as np
from .a_XMolLinesEditor import *
from .a_XFileMoleculeHistogram import *
import os.path
import webbrowser
import sys
from ._shared import *
import a99
import pyfant

__all__ = ["XFileMolecules"]


NUM_PLOTS = len(SOL_HEADERS_PLOT)-1  # -1 because whe "lambda" does not have its plot
MAX_NUM_BINS = 500

class XFileMolecules(QMainWindow):

    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)

        self.f = None  # FileMolecules object
        self.flag_sort = False
        self.mol_index = None
        self.sol_index = None
        self.mol = None
        self.sol = None
        self.save_dir = "."
        self.flag_changed = False
        self.form_lines = None

        # Information about the plots
        self.marker_row = None  # points into current set-of-lines, self.sol
        self.plot_info = [PlotInfo() for _ in range(NUM_PLOTS)]
        self.set_flag_plot(SOL_ATTR_NAMES.index("sj")-1, True)
        self.set_flag_plot(SOL_ATTR_NAMES.index("jj")-1, True)

        # ** tab "General file info"
        a = self.plainTextEditFileInfo = QPlainTextEdit()
        a.setFont(a99.MONO_FONT)


        # ** "Molecules browser"

        # ** ** left of splitter
        self.labelMol = QLabel('Molecules list (Ctrl+1)')
        a = self.listWidgetMol = QListWidget()
        a.currentRowChanged.connect(self.on_listWidgetMol_currentRowChanged)
        # a.setEditTriggers(QAbstractItemView.DoubleClicked)
        # a.setEditTriggers(QAbstractItemView.AllEditTriggers)
        a.setContextMenuPolicy(Qt.CustomContextMenu)
        a.customContextMenuRequested.connect(self.on_listWidgetMol_customContextMenuRequested)
        a.itemDoubleClicked.connect(self.on_listWidgetMol_doubleClicked)
        a.installEventFilter(self)

        l = self.layoutMol = QVBoxLayout()
        a99.set_margin(l, 0)
        l.setSpacing(1)
        l.addWidget(self.labelMol)
        l.addWidget(self.listWidgetMol)

        a = self.widgetMol = QWidget()
        a.setLayout(self.layoutMol)


        # ** ** right of splitter

        # ** ** ** tab "Molecule info"

        # ** ** ** ** Toolbar with molecule-specific resources

        l2 = self.labelSpinBoxBins = QLabel("&Bins")
        sb = self.spinBoxBins = QSpinBox()
        sb.setMaximum(MAX_NUM_BINS)
        sb.setValue(50)
        l2.setBuddy(sb)

        am = self.buttonWavelengthHistogram = QPushButton("Wavelength histogram")
        am.clicked.connect(self.on_buttonWavelengthHistogram_clicked)
        am.setToolTip("Plots histogram that gives a general profile of where the lines are wavelength-wise")

        self.spacer1625 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        l0 = self.layoutMolToolbar = QHBoxLayout()
        l0.addWidget(l2)
        l0.addWidget(sb)
        l0.addWidget(am)
        l0.addItem(self.spacer1625)
        a99.set_margin(l0, 1)
        a = self.widgetMolToolbar = QWidget()
        a.setLayout(l0)
        a.setFixedHeight(40)

        self.labelMolInfo = QLabel('Molecule summary report')
        a = self.plainTextEditMolInfo = QPlainTextEdit()
        a.setFont(a99.MONO_FONT)

        l = self.layoutMolInfo = QVBoxLayout()
        a99.set_margin(l, 0)
        l.setSpacing(1)
        l.addWidget(self.widgetMolToolbar)
        l.addWidget(self.labelMolInfo)
        l.addWidget(self.plainTextEditMolInfo)

        a = self.widgetMolInfo = QWidget()
        a.setLayout(self.layoutMolInfo)

        # ** ** ** tab "Sets of lines (Alt+L)"

        # ** ** ** ** left

        # P = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        self.labelSol = QLabel('Sets of lines (Ctrl+2)')

        a = self.listWidgetSol = QListWidget()
        a.setFont(a99.MONO_FONT)
        # a.setFixedWidth(100)
        a.currentRowChanged.connect(self.on_listWidgetSol_currentRowChanged)
        a.setContextMenuPolicy(Qt.CustomContextMenu)
        a.customContextMenuRequested.connect(self.on_listWidgetSol_customContextMenuRequested)
        a.itemDoubleClicked.connect(self.on_listWidgetSol_doubleClicked)
        a.installEventFilter(self)

        l = self.layoutSol = QVBoxLayout()
        a99.set_margin(l, 0)
        l.setSpacing(1)
        l.addWidget(self.labelSol)
        l.addWidget(self.listWidgetSol)

        a = self.widgetSol = QWidget()
        a.setLayout(self.layoutSol)

        # ** ** ** ** right

        # ** ** ** ** ** Toolbar above plot

        am = self.buttonSort = QPushButton("Sort wave (Alt+&W)")
        am.clicked.connect(self.on_buttonSort_clicked)
        am.setCheckable(True)
        am.setToolTip("Sort spectral lines in ascending order of wavelength")
        if self.flag_sort:
            am.setChecked(True)

        # adds checkable buttons sj, jj
        self.plot_buttons = bb = []
        for i in range(NUM_PLOTS):  # sj, jj etc
            s = SOL_HEADERS[i+1]
            b = QPushButton("%s (Alt+&%d)" % (s, i+1))
            b.clicked.connect(self.on_button_plot_clicked)
            b.setCheckable(True)
            if self.flag_plot(i):
                b.setChecked(True)
            bb.append(b)

        a11 = self.buttonEditLines = QPushButton("Edit lines (Ctrl+3)")
        a11.clicked.connect(self.on_buttonEditLines_clicked)
        a2 = self.labelNumLines = QLabel("--")
        a3 = self.spacer0 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        l0 = self.layoutSolToolbar = QHBoxLayout()
        l0.addWidget(am)
        for b in bb:
            l0.addWidget(b)
        l0.addWidget(a11)
        l0.addWidget(a2)
        l0.addItem(a3)
        a99.set_margin(l0, 1)
        a = self.widgetSolToolbar = QWidget()
        a.setLayout(l0)
        a.setFixedHeight(40)

        # ** ** ** ** ** Set-of-lines info + Plot TABS

        # ** ** ** ** ** ** tab "Set-of-lines info"
        a = self.plainTextEditSolInfo = QPlainTextEdit()
        a.setFont(a99.MONO_FONT)

        # ** ** ** ** ** ** Plot tab

        # http://stackoverflow.com/questions/12459811
        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.canvas.mpl_connect('button_press_event', self.on_plot_click)
        self.toolbar = NavigationToolbar2QT(self.canvas, self)
        layout = QVBoxLayout()
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        a99.set_margin(layout, 0)

        a = self.widgetSolPlot = QWidget()
        a.setLayout(layout)

        l1 = self.layoutSolPlot = QVBoxLayout()
        l1.addWidget(self.widgetSolToolbar)
        l1.addWidget(self.widgetSolPlot)
        a99.set_margin(l1, 0)

        a = self.widgetSolPlot = QWidget()
        a.setLayout(l1)


        # ** ** ** ** **  Set-of-lines info + Plot TABS
        a = self.tabWidgetSol = QTabWidget(self)
        a.addTab(self.plainTextEditSolInfo, "Set-of-lines Info (Alt+&N)")
        a.addTab(self.widgetSolPlot, "Set-of-lines plots (Alt+&P)")
        a.setCurrentIndex(1)
        a.setFont(a99.MONO_FONT)

        # ** ** ** ** ** splitter: (list of set-of-lines) | (plot)
        a = self.splitterSol = QSplitter(Qt.Horizontal)
        a.addWidget(self.widgetSol)
        a.addWidget(self.tabWidgetSol)
        a.setStretchFactor(0, 2)
        a.setStretchFactor(1, 10)

        a = self.tabWidgetMol = QTabWidget()
        a.addTab(self.widgetMolInfo, "Molecule info (Alt+&M)")
        a.addTab(self.splitterSol, "Sets of lines (Alt+&L)")
        a.setCurrentIndex(1)
        a.setFont(a99.MONO_FONT)

        # ** splitter: (list of molecules) | (molecules tab widget)
        a = self.splitterMol = QSplitter(Qt.Horizontal)
        a.addWidget(self.widgetMol)
        a.addWidget(self.tabWidgetMol)
        a.setStretchFactor(0, 2)
        a.setStretchFactor(1, 10)

        # tab "File" (main tab widget
        a = self.tabWidgetFile = QTabWidget(self)
        a.addTab(self.plainTextEditFileInfo, "General File Info (Alt+&I)")
        a.addTab(self.splitterMol, "Molecules Browser (Alt+&B)")
        a.setCurrentIndex(1)
        a.setFont(a99.MONO_FONT)


        # * # * # * # * # * # * # *
        # Now the menu bar

        # self.menubar = QMenuBar(self)
        # self.menubar.setGeometry(QRect(0, 0, 772, 18))
        # self.menubar.setObjectName(_fromUtf8("menubar"))
        b = self.menuBar()
        m = self.menu_file = b.addMenu("&File")
        self.act_save = ac = m.addAction("&Save")
        ac.setShortcut("Ctrl+S")
        ac.triggered.connect(self.on_save)
        self.act_save_as = ac = m.addAction("Save &as...")
        ac.setShortcut("Ctrl+Shift+S")
        ac.triggered.connect(self.on_save_as)
        m.addSeparator()
        ac = m.addAction("&Quit")
        ac.setShortcut("Ctrl+Q")
        ac.triggered.connect(self.close)

        # * # * # * # * # * # * # *
        # Final adjustments

        self.setCentralWidget(self.tabWidgetFile)
        a99.place_left_top(self)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #

    def keyPressEvent(self, ev):
        if ev.modifiers() == Qt.ControlModifier:
            k = ev.key()
            if k == Qt.Key_1:
                self.listWidgetMol.setFocus()
            elif k == Qt.Key_2:
                self.listWidgetSol.setFocus()
            elif k == Qt.Key_3:
                self.on_buttonEditLines_clicked()

    def eventFilter(self, source, event):
        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Return:
                if source == self.listWidgetMol:
                    self.edit_mol()
                    return True
                if source == self.listWidgetSol:
                    self.edit_sol()
                    return True
            if event.key() == Qt.Key_Delete:
                if source == self.listWidgetMol:
                    self.delete_mol()
                    return True
                elif source == self.listWidgetSol:
                    self.delete_sol()
                    return True
        return False

    def closeEvent(self, event):
        if self.flag_changed:
            # http://straightedgelinux.com/blog/python/html/pyqtxt.html
            r = QMessageBox.question(self,
                        "About to exit",
                        "File \"%s\" has unsaved changes. Save now?" % self.f.filename,
                        QMessageBox.Yes | QMessageBox.No|
                        QMessageBox.Cancel)
            if r == QMessageBox.Cancel:
                event.ignore()
            else:
                if r == QMessageBox.No:
                    pass
                elif r == QMessageBox.Yes:
                    try:
                        self.save()
                    except:
                        # In case of error saving file, will not exit the program
                        event.ignore()
        if event.isAccepted():
            self.close_editor()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots

    def on_help(self, _):
        try:
            base_dir = os.path.dirname(sys.argv[0])
            webbrowser.open_new(os.path.join(base_dir, "mled.html"))
            a99.show_message("Help file mled.html was opened in web browser.")
        except Exception as e:
            a99.show_error(str(e))

    def on_save(self, _):
        self.disable_save_actions()
        try:
            self.save()
        except Exception as e:
            a99.show_error(str(e))
        finally:
            self.enable_save_actions()

    def on_save_as(self, _):
        self.disable_save_actions()
        try:
            if self.f:
                new_filename = QFileDialog.getSaveFileName(self, "Save file",
                 self.save_dir, "*.dat")[0]
                if new_filename:
                    self.save_dir, _ = os.path.split(str(new_filename))
                    self.save_as(new_filename)
        except Exception as e:
            a99.show_error(str(e))
        finally:
            self.enable_save_actions()

    def on_listWidgetMol_currentRowChanged(self, row):
        if row > -1:
            self.set_molecule(row)

    def on_listWidgetSol_currentRowChanged(self, row):
        if row > -1:
            self.set_sol(row)

    def on_button_plot_clicked(self):
        # This may be triggered by any of the plot buttons
        button = self.sender()
        idx = self.plot_buttons.index(button)
        self.set_flag_plot(idx, button.isChecked())
        self.plot_lines()

    def on_buttonSort_clicked(self):
        self.flag_sort = self.buttonSort.isChecked()
        self.plot_lines()

    def on_listWidgetMol_customContextMenuRequested(self, position):
        menu = QMenu()
        a_edit = menu.addAction("&Edit")
        a_delete = menu.addAction("&Delete")
        action = menu.exec_(self.listWidgetMol.mapToGlobal(position))
        if action == a_edit:
            self.edit_mol()
        elif action == a_delete:
            self.delete_mol()

    def on_listWidgetSol_customContextMenuRequested(self, position):
        menu = QMenu()
        a_edit = menu.addAction("&Edit")
        a_delete = menu.addAction("&Delete")
        action = menu.exec_(self.listWidgetSol.mapToGlobal(position))
        if action == a_edit:
            self.edit_sol()
        elif action == a_delete:
            self.delete_sol()

    def on_listWidgetMol_doubleClicked(self):
        self.edit_mol()

    def on_listWidgetSol_doubleClicked(self):
        self.edit_sol()

    def on_buttonWavelengthHistogram_clicked(self):
        f = plt.figure()
        pyfant.plot_molhistogram(self.mol, num_bins=self.spinBoxBins.value())
        plt.title(self.mol.description)
        f.show()
        # form = XFileMoleculeHistogram(self.mol)
        # form.show()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #

    def load(self, f):
        """Loads file into GUI."""
        assert isinstance(f, pyfant.FileMolecules)

        self.f = f

        self.plainTextEditFileInfo.setPlainText(str(f))

        for m in f.molecules:
            assert isinstance(m, pyfant.Molecule)
            item = QListWidgetItem(self.get_mol_string(m))
            # not going to allow editing yet item.setFlags(item.flags() | Qt.ItemIsEditable)
            self.listWidgetMol.addItem(item)

        if len(f.molecules) > 0:
            self.listWidgetMol.setCurrentRow(0)

        self.flag_changed = False
        self.update_window_title()

    def save(self):
        if self.f:
            self.f.save_as()
            self.flag_changed = False
            self.update_window_title()

    def save_as(self, filename):
        if self.f:
            self.f.save_as(filename)
            self.flag_changed = False
            self.update_window_title()

    def set_molecule(self, i):
        self.mol_index = i
        m = self.mol = self.f.molecules[i] if len(self.f.molecules) > i else None

        self.update_mol_info()

        w = self.listWidgetSol
        w.clear()

        if m is not None:
            for i, sol in enumerate(m.sol):
                item = QListWidgetItem(self.get_sol_string(i, sol))
                w.addItem(item)

            if len(m) > 0:
                w.setCurrentRow(0)

    def set_sol(self, j):
        # TODO what if molecule has no set of lines (SOL)? Actually I can't allow molecule to have 0 SOLs
        j = min(j, len(self.mol)-1)

        self.marker_row = None  # no longer valid, whatever it was
        self.sol_index = j
        self.sol = self.f.molecules[self.mol_index].sol[j]
        self.update_sol_info()
        self.plot_lines()
        self.set_editor_sol()

    def plot_lines(self):
        try:
            self.clear_markers()
            o = self.sol
            if o is not None:
                self.figure.clear()

                n = sum([info.flag for info in self.plot_info])  # number of subplots (0, 1 or 2)
                # map to reuse plotting routine, contains what differs between each plot

                map_ = [(SOL_HEADERS_PLOT[i], o.__getattribute__(SOL_ATTR_NAMES[i]))
                        for i in range(1, len(SOL_HEADERS_PLOT))]

                i_subplot = 1
                refaxis = None
                for i in range(len(map_)):
                    y_label = map_[i][0]
                    pi = self.plot_info[i]
                    pi.y_vector = _y = map_[i][1]

                    if pi.flag:
                        if not self.flag_sort:
                            x = o.lmbdam
                            y = _y
                        else:
                            _x = np.array(o.lmbdam)
                            _y = np.array(_y)
                            ii = np.argsort(_x)
                            x = _x[ii]
                            y = _y[ii]

                        a99.format_BLB()

                        self.figure.add_subplot(n, 1, i_subplot, sharex=refaxis)
                        pi.axis = ax = self.figure.gca()
                        if i == 0: refaxis = ax  # for sharex next i-iteration
                        ax.clear()
                        ax.plot(x, y, 'k'+('' if len(x) > 1 else 'x'))
                        ax.set_xlabel('Wavelength ($\AA$)')
                        ax.set_ylabel(y_label)

                        # x-limits
                        xmin, xmax = min(x), max(x)
                        k = .02*(xmax-xmin)
                        ax.set_xlim([xmin-k, xmax+k])

                        # y-limits
                        ymin, ymax = min(y), max(y)
                        k = .02*(ymax-ymin)
                        ax.set_ylim([ymin-k, ymax+k])

                        i_subplot += 1

                if i_subplot > 1:
                    plt.tight_layout()

                self.canvas.draw()
                self.draw_markers()
        except Exception as e:
            a99.show_error("Error drawing plots: {}".format(str(e)))


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #

    def edit_mol(self):
        obj = self.mol
        if obj is None:
            return
        item = self.listWidgetMol.currentItem()
        r, form = a99.show_edit_form(obj,
            ["description", "symbols", "fe", "do", "mm", "am", "bm", "ua", "ub", "te", "cro", "s"],
            item.text())
        flag_changed = False
        if r == QDialog.Accepted:
            kwargs = form.get_kwargs()
            for name, value in kwargs.items():
                orig = obj.__getattribute__(name)
                if orig != value:
                    obj.__setattr__(name, value)
                    flag_changed = True
        if flag_changed:
            self.flag_changed = True
            item.setText(self.get_mol_string(obj))
            # item.setStyleSheet("selected:active{background: yellow}")
            # item.setTextColor(QColor(255, 0, 0))
            # item.setBackgroundColor(QColor(255, 0, 0))
            self.update_mol_info()
            self.update_window_title()

    def delete_mol(self):
        f = self.f
        if f is None:
            return
        i = self.listWidgetMol.currentRow()
        if i >= len(f.molecules):
            return
        del self.f.molecules[i]
        self.listWidgetMol.takeItem(i)
        self.flag_changed = True
        self.update_window_title()


    def edit_sol(self):
        obj = self.sol
        if obj is None:
            return
        item = self.listWidgetSol.currentItem()
        r, form = a99.show_edit_form(self.sol, ["vl", "v2l", "qqv", "ggv", "bbv", "ddv", "fact"],
                                    item.text())
        flag_changed = False
        if r == QDialog.Accepted:
            kwargs = form.get_kwargs()
            for name, value in kwargs.items():
                orig = obj.__getattribute__(name)
                if orig != value:
                    obj.__setattr__(name, value)
                    flag_changed = True
        if flag_changed:
            self.flag_changed = True
            item.setText(self.get_sol_string(self.listWidgetSol.currentRow(), obj))
            # item.setTextColor(QColor(255, 0, 0))
            self.update_sol_info()
            self.update_window_title()

    def delete_sol(self):
        sol = self.sol
        if sol is None:
            return
        i = self.listWidgetSol.currentRow()
        if i >= len(self.mol.sol):
            return
        del self.mol.sol[i]
        self.listWidgetSol.takeItem(i)
        self.flag_changed = True
        self.update_window_title()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #

    def update_mol_info(self):
        """Makes report for the current molecule."""
        m = self.mol
        s = str(m)
        # No need to repeat this information, already shown in listWidgetSol
        # s += '\n\n' \
        # 'Sets of lines\n' \
        #      '-------------\n' \
        #      ' # Number of lines\n'
        # for i in range(len(m)):
        #     s += '%2d %15d\n' % (i+1, len(m.lmbdam[i]))
        self.plainTextEditMolInfo.setPlainText(s)

    def update_sol_info(self):
        """Makes report for the current set-of-lines."""
        o = self.sol
        s = str(o)
        self.plainTextEditSolInfo.setPlainText(s)
        n = len(o)
        self.labelNumLines.setText('Number of lines: %d' % (n,))

    def update_window_title(self):
        self.setWindowTitle("mled - %s" % (self.f.filename+("" if not self.flag_changed
                                                            else " (changed)"),))

    def enable_save_actions(self):
        self.act_save.setEnabled(True)
        self.act_save_as.setEnabled(True)

    def disable_save_actions(self):
        self.act_save.setEnabled(False)
        self.act_save_as.setEnabled(False)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Routines related to the lines editor

    def on_buttonEditLines_clicked(self):
        if self.form_lines is None:
            f = self.form_lines = XMolLinesEditor(self)
            f.show()
            self.set_editor_sol()
        self.plot_lines()

    def MolLinesEditor_closing(self):
        """Called by the molecular lines editor to notify that it is closing."""
        self.form_lines = None
        self.marker_row = None
        self.plot_lines()  # to remove the "X" from the plot

    def MolLinesEditor_current_row_changed(self, currentRow):
        """Called by the molecular lines editor to notify that the current row has changed."""
        self.set_marker_row(currentRow)

    def MolLinesEditor_cell_changed(self, row, column, value):
        """Called by the molecular lines editor to notify that a value has changed."""

        try:
            attr_name = SOL_ATTR_NAMES[column]
            v = self.sol.__getattribute__(attr_name)
            if v[row] != value:
                v[row] = value
                self.flag_changed = True
                self.plot_lines()
                self.update_window_title()
        except Exception as e:
            a99.get_python_logger().exception("XFileMolecules.MolLinesEditor_cell_changed() raised")

    def set_editor_sol(self):
        """Sets the set-of-lines of the editor."""
        if self.sol is not None and self.form_lines is not None:
            self.form_lines.set_sol(self.sol, "Set-of-lines: "+
                                    self.listWidgetSol.currentItem().text())

    def close_editor(self):
        if self.form_lines is not None:
            self.form_lines.close()

    # Controlling plots and markers for current row

    def set_marker_row(self, i):
        self.clear_markers()
        self.marker_row = i
        self.draw_markers()

    def clear_markers(self):
        for o in self.plot_info:
            if o.mpl_obj:
                a99.remove_line(o.mpl_obj)
                o.mpl_obj = None

    def draw_markers(self):
        self.clear_markers()
        if self.marker_row is not None and any([o.flag for o in self.plot_info]):
            i = self.marker_row
            lambda_ = self.sol.lmbdam[i]
            for o in self.plot_info:
                if o.flag:
                    # http://stackoverflow.com/questions/22172565/matplotlib-make-plus-sign-thicker
                    o.mpl_obj = o.axis.plot([lambda_], [o.y_vector[i]], 'xr', mew=2, ms=10)
            self.canvas.draw()

    def flag_plot(self, idx):
        return self.plot_info[idx].flag

    def set_flag_plot(self, idx, x):
        self.plot_info[idx].flag = x

    def on_plot_click(self, event):
        lambda_ = event.xdata
        if lambda_ is not None and self.form_lines is not None:
            idx = a99.index_nearest(self.sol.lmbdam, lambda_)
            self.form_lines.set_row(idx)
            # self.set_marker_row(idx)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #

    @staticmethod
    def get_mol_string(m):
        return m.description

    @staticmethod
    def get_sol_string(index, sol):
        return "%3d (%2d, %2s) %7s" % (index+1, sol.vl, sol.v2l, '(%d)' % len(sol))



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
import pyfant
import f311

__all__ = ["XFileMolConsts"]


class XFileMolConsts(f311.XFileMainWindowBase):
    def __init__(self, *args, moldb=None, **kwargs):
        f311.XFileMainWindowBase.__init__(self, *args, **kwargs)
        if moldb is not None:
            self.me.set_moldb(moldb)

    def set_moldb(self, moldb):
        self.me.set_moldb(moldb)

    def _add_stuff(self):
        me = self.me = pyfant.WFileMolConsts(self)
        self.tabWidget.addTab(me, "")

        _VVV = pyfant.FileMolConsts.description
        self.pages.append(f311.MyPage(
         text_tab="{}".format(_VVV),
         text_saveas="Save %s as..." % _VVV,
         text_load="Load %s" % _VVV,
         cls_save=pyfant.FileMolConsts, clss_load=(pyfant.FileMolConsts,), wild="*.py",
         editor=me
        ))

        self.setWindowTitle("Molecular constants editor")

    def wants_auto(self):
        idx = self.w_source.index
        filename = None
        if idx == 0:
            lines = self.w_hitran.data
            if lines:
                filename = "{}.dat".format(lines["header"]["table_name"])

        if filename is None:
            # Default
            filename = a99.new_filename("mol", "dat")

        self.w_out.value = filename


    # # # Override
    # #   ========
    #
    # def _on_me_changed(self):
    #     """Overriden to commit automatically. "(changed)" will not appear
    #
    #     Changes should be committed by the method that executed queries, but it is so easy to
    #     reinforce this... just in case.
    #     """
    #     index = self._get_tab_index()
    #     if index == 0:
    #         self.me.f.get_conn().commit()  # Just in case
    #     else:
    #         self.flags_changed[index] = True
    #         self._update_tab_texts()



#!/usr/bin/env python
"""
Copies stellar data files (such as main.dat, abonds.dat, dissoc.dat) to local directory

examples of usage:
  > copy-star.py
  (displays menu)

  > copy-star.py arcturus
  ("arcturus" is the name of a subdirectory of PFANT/data)

  > copy-star.py -p /home/user/pfant-common-data
  (use option "-p" to specify path)

  > copy-star.py -l
  (lists subdirectories of PFANT/data , doesn't copy anything)

"""
import argparse
import logging
import os.path
import sys
import glob
import shutil
import pyfant
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    flag_menu = len(sys.argv) == 1  # will display menu if no command-line arguments

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=a99.SmartFormatter
    )
    parser.add_argument('-l', '--list', action='store_true',
      help='lists subdirectories of '+pyfant.get_pfant_data_path())
    parser.add_argument('-p', '--path', action='store_true',
      help='system path mode')
    parser.add_argument('directory', type=str, nargs="?",
     help='name of directory (either a subdirectory of PFANT/data or the path '
          'to a valid system directory (see modes of operation)')

    args = parser.parse_args()


    # "-l" mode
    if args.list:
        print("\n".join(a99.format_h1("Subdirectories of '{}'".format(pyfant.get_pfant_data_path()))))
        for dirname in pyfant.get_pfant_data_subdirs():
            print(dirname)
        sys.exit()
        
    # figures out the path to directory (dir_)
    if flag_menu:
        print("\nLooking into directory '{}'...".format(pyfant.get_pfant_data_path()))
        dirnames = pyfant.get_pfant_star_subdirs()
        dirnames.sort()
        choice = a99.menu("Choose a star", ["{}{}".format(x[0].upper(), x[1:]) for x in dirnames],
                      cancel_label="quit", flag_allow_empty=True, flag_cancel=False)
        if choice is None or choice <= 0:
            sys.exit()
        dir_ = os.path.join(pyfant.get_pfant_data_path(), dirnames[choice-1])
    else:
        if args.path:
            dir_ = args.directory
        else:
            dir_ = os.path.join(pyfant.get_pfant_data_path(), args.directory)
            
    if not os.path.isdir(dir_):
        print("'%s' is not a directory" % dir_)
        sys.exit(-1)
    pyfant.copy_star(dir_)

    # print("\n")
    # print("\n".join(a99.format_box("Attention")))
    # print("File 'dissoc.dat' may require revision!\n\n"
    #       "You may delete this file to assume atomic abundances in 'abonds.dat' for the molecules.\n\n"
    #       "(this message should be deleted when all files 'dissoc.dat' provided with PFANT\n"
    #       "finish being reviewed)\n\n"
    #       "(2017-11-28)\n")



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from f311 import hapi
import os
import datetime
from collections import OrderedDict
import f311
import pyfant


__all__ = ["XConvMol"]


# Spacing for grid layouts
_LAYSP_GRID = 4
# Margin for grid layouts
_LAYMN_GRID = 0
# Spacing for vertical layouts
_LAYSP_V = 10
# Margin for vertical layouts ((caption, combobox); grid)
_LAYMN_V = 6


class _DataSource(a99.AttrsPart):
    """Represents a data source for molecular lines"""

    def __init__(self, name):
        a99.AttrsPart.__init__(self)
        self.name = name
        self.widget = None

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.name)


# This defines the order of the panels
_NAMES = ["HITRAN", "VALD3", "Kurucz", "Plez", ]
_SOURCES = OrderedDict([[name, _DataSource(name)] for name in _NAMES])


class _WSource(a99.WBase):
    """Lists sources for molecular lines data"""

    @property
    def index(self):
        return self._get_index()

    @index.setter
    def index(self, x):
        self._buttons[x].setChecked(True)

    @property
    def sourcename(self):
        source = self._get_source()
        if source is None:
            return None
        return source.name

    @sourcename.setter
    def sourcename(self, value):
        index = -1
        if value is not None:
            index = _NAMES.index(value)
        for i, b in enumerate(self._buttons):
            b.setChecked(i == index)
        self._check_index_changed(False)

    @property
    def source(self):
        return self._get_source()


    # Emitted if index changed by user, not programatically
    index_changed = pyqtSignal()

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._buttons = []
        self._last_index = -1

        lw = QVBoxLayout()
        self.setLayout(lw)

        for ds in _SOURCES.values():
            b = self.keep_ref(QRadioButton(ds.name))

            b.clicked.connect(self._button_clicked)
            self._buttons.append(b)
            lw.addWidget(b)


    def _get_source(self):
        """Returns _DataSource object or None"""
        i = self._get_index()
        if i < 0:
            return None
        return _SOURCES[_NAMES[i]]

    def _button_clicked(self):
        self._check_index_changed()

    def _check_index_changed(self, flag_emit=True):
        i = self._get_index()
        if i != self._last_index:
            self._last_index = i
            if flag_emit:
                self.index_changed.emit()

    def _get_index(self):
        for i, b in enumerate(self._buttons):
            assert isinstance(b, QRadioButton)
            if b.isChecked():
                return i
        return -1


class _WSelectSaveFile(a99.WBase):
    @property
    def value(self):
        return self._get_value()

    @value.setter
    def value(self, x):
        if x is None:
            x = ""
        self.edit.setText(x)

    @property
    def flag_valid(self):
        return self._flag_valid

    @flag_valid.setter
    def flag_valid(self, value):
        self._flag_valid = value
        self._update_gui()

    # Emitted whenever the valu property changes **to a valid value**
    wants_autofilename = pyqtSignal()
    # Emitted whenever text changes
    changed = pyqtSignal()

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._last_value = None
        self._flag_valid = None

        self._type = None
        self.dialog_title = "Save output as"
        self.dialog_path = "."
        self.dialog_wild = "*.*;;*.dat"

        lw = QHBoxLayout()
        self.setLayout(lw)

        t = self.label = self.keep_ref(QLabel("Save output as"))
        lw.addWidget(t)

        e = self.edit = QLineEdit()
        e.textEdited.connect(self.edit_changed)
        t.setBuddy(e)
        lw.addWidget(e)
        # e.setReadOnly(True)

        b = self.button_autofilename = QToolButton()
        lw.addWidget(b)
        b.clicked.connect(self.wants_autofilename)
        b.setIcon(a99.get_icon("leaf-plant"))
        b.setToolTip("Make up file name")
        b.setFixedWidth(30)

        b = self.button = QToolButton()
        lw.addWidget(b)
        b.clicked.connect(self.on_button_clicked)
        b.setIcon(a99.get_icon("document-save"))
        b.setToolTip("Choose file name to save as")
        b.setFixedWidth(30)

        # Forces paint red if invalid
        self._update_gui()

    def on_button_clicked(self, _):
        self._on_button_clicked()

    def edit_changed(self):
        self.changed.emit()

    def _update_gui(self):
        a99.style_widget_valid(self.edit, self._flag_valid)

    def _on_button_clicked(self):
        path_ = self.edit.text().strip()
        if len(path_) == 0:
            path_ = self.dialog_path
        res = QFileDialog.getSaveFileName(self, self.dialog_title, path_, self.dialog_wild)[0]
        if res:
            # res = res[0]
            self.edit.setText(res)
            self.changed.emit()
            self.dialog_path = res

    # def _wanna_emit(self):
    #     value_now = self._get_value()
    #     if value_now != self._last_value:
    #         self._last_value = value_now
    #         self.valueChanged.emit()

    def _get_value(self):
        return self.edit.text().strip()


class _WHitranPanel(a99.WBase):

    @property
    def data(self):
        """
        Returns value in hapi.LOCAL_TABLE_CACHE, or None. This variable is a dictionary

        See hapi.py for structure of this variable
        """
        idx = self.tableWidget.currentRow()
        if idx < 0:
            return None
        return hapi.LOCAL_TABLE_CACHE[self.tableWidget.item(idx, 0).text()]

    def __init__(self, parent):
        a99.WBase.__init__(self, parent.parent_form)

        self.w_conv = parent

        self._flag_populating = False

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel("HITRAN")))


        lg = QGridLayout()
        lw.addLayout(lg)

        a = self.keep_ref(QLabel("HITRAN 'data cache' directory"))
        w = self.w_dir = a99.WSelectDir(self.parent_form)
        a.setBuddy(w)
        w.changed.connect(self.dir_changed)
        lg.addWidget(a, 0, 0)
        lg.addWidget(w, 0, 1)

        a = self.tableWidget = QTableWidget()
        lw.addWidget(a)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setFont(a99.MONO_FONT)
        a.setAlternatingRowColors(True)
        a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)

        # forces populate table with 'Python HITRAN API data cache' in local directory
        self.dir_changed()

    def on_tableWidget_currentCellChanged(self, curx, cury, prevx, prevy):
        pass

    def dir_changed(self):
        self._populate()


    def _populate(self):
        self._flag_populating = True
        try:
            # Changed HAPI working directory
            hapi.VARIABLES['BACKEND_DATABASE_NAME'] = self.w_dir.value
            # Loads all molecular lines data to memory
            hapi.loadCache()

            # Discounts "sampletab" table from HAPI cache, hence the "-1" below
            nr, nc = len(hapi.LOCAL_TABLE_CACHE)-1, 2
            t = self.tableWidget
            a99.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(["Table filename (.data & .header)", "Number of spectral lines"])

            i = 0
            for h, (name, data) in enumerate(hapi.LOCAL_TABLE_CACHE.items()):
                if name == "sampletab":
                    continue

                header = data["header"]

                item = QTableWidgetItem(name)
                t.setItem(i, 0, item)

                item = QTableWidgetItem(str(header["number_of_rows"]))
                t.setItem(i, 1, item)

                i += 1

            t.resizeColumnsToContents()

            # self._data = rows
            #
            # if restore_mode == "formula":
            #     if curr_row:
            #         self._find_formula(curr_row["formula"])
            # elif restore_mode == "index":
            #     if -1 < curr_idx < t.rowCount():
            #         t.setCurrentCell(curr_idx, 0)
        finally:
            self._flag_populating = False
            # self._wanna_emit_id_changed()


# TODO other panels must comply with Kurucz
class _WVald3Panel(a99.WBase):
    """
    This panel allows to load a Vald3 file and browse through its species (molecules only)

    The goal is to choose one molecule
    """

    @property
    def data(self):
        """
        Returns FileVald3 with a single species"""

        idx = self.tableWidget.currentRow()
        if idx < 0:
            return None

        f = pyfant.FileVald3()
        f.speciess = [self._f.speciess[idx]]
        return f

    @property
    def is_molecule(self):
        data = self.data
        return data is not None and data.speciess[0].formula not in pyfant.symbols

    def __init__(self, parent):
        a99.WBase.__init__(self, parent.parent_form)

        self.w_conv = parent

        self._flag_populating = False
        self._f = None  # FileVald3

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel("VALD3")))

        lg = QGridLayout()
        lw.addLayout(lg)

        a = self.keep_ref(QLabel("VALD3 extended-format file"))
        w = self.w_file = a99.WSelectFile(self.parent_form)
        a.setBuddy(w)
        w.changed.connect(self.file_changed)
        # TODO I bumped into this and removed it lw.addWidget(w)
        lg.addWidget(a, 0, 0)
        lg.addWidget(w, 0, 1)


        a = self.tableWidget = QTableWidget()
        lw.addWidget(a)
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setFont(a99.MONO_FONT)
        a.setAlternatingRowColors(True)
        a.currentCellChanged.connect(self.on_tableWidget_currentCellChanged)

        l = self.label_warning = QLabel()
        l.setStyleSheet("QLabel {{color: {}}}".format(a99.COLOR_WARNING))
        lw.addWidget(l)

        # forces populate table with 'Python HITRAN API data cache' in local directory
        # self.file_changed()

    def on_tableWidget_currentCellChanged(self, curx, cury, prevx, prevy):
        self.label_warning.setText("Need a molecule, not atom"
                                   if self.data is not None and not self.is_molecule else "")

    def file_changed(self):
        self._populate()


    def _populate(self):
        self._flag_populating = True
        try:
            f = self._f = pyfant.FileVald3()
            f.load(self.w_file.value)

            nr, nc = len(f), 3
            t = self.tableWidget
            a99.reset_table_widget(t, nr, nc)
            t.setHorizontalHeaderLabels(["VALD3 species", "Number of spectral lines", "Atom/Molecule"])

            for i, species in enumerate(f):
                item = QTableWidgetItem(str(species))
                t.setItem(i, 0, item)
                item = QTableWidgetItem(str(len(species)))
                t.setItem(i, 1, item)
                item = QTableWidgetItem("Atom" if species.formula in pyfant.symbols else "Molecule")
                t.setItem(i, 2, item)

            t.resizeColumnsToContents()

        except Exception as e:
            self._f = None
            self.add_log_error("Error reading contents of file '{}': '{}'".format(self.w_file.value, a99.str_exc(e)), True)
            raise

        else:
            self.clear_log()

        finally:
            self._flag_populating = False
            # self._wanna_emit_id_changed()


class _WKuruczPanel(a99.WBase):
    """
    This panel allows to load a Kurucz molecular lines file
    """

    @property
    def data(self):
        """Returns FileKuruczMoleculeBase or None"""
        return self._flines

    @property
    def iso(self):
        return self._get_iso()

    @iso.setter
    def iso(self, value):
        self._set_iso(value)

    def __init__(self, parent):
        a99.WBase.__init__(self, parent.parent_form)

        self.w_conv = parent

        self._flines = None  # FileKuruczMoleculeBase
        # list of integers, filled when file is loaded
        self._isotopes = []

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel("Kurucz")))

        lg = QGridLayout()
        lw.addLayout(lg)

        i_row = 0
        a = self.keep_ref(QLabel("Kurucz molecular lines file"))
        w = self.w_file = a99.WSelectFile(self.parent_form)
        w.changed.connect(self._file_changed)
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Isotope"))
        w = self.combobox_isotope = QComboBox()
        w.currentIndexChanged.connect(self.changed)
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        lw.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

    def _set_iso(self, value):
        if value is None:
            idx = 0
        else:
            try:
                idx = self._isotopes.index(value)+1
            except ValueError:
                idx = 0
        cb = self.combobox_isotope
        if cb.count() > 0:
            save = cb.blockSignals(True)
            cb.setCurrentIndex(idx)
            cb.blockSignals(save)

    def _get_iso(self):
        idx = self.combobox_isotope.currentIndex()
        if idx <= 0:
            return None
        return self._isotopes[idx - 1]

    def _file_changed(self):

        f = None
        try:
            f = self._load_lines()

        except Exception as e:
            self._flines = None
            self._isotopes = []
            msg = "Error reading contents of file '{}': '{}'".format(self.w_file.value,
                                                                              a99.str_exc(e))
            self.add_log_error(msg, True)
            a99.get_python_logger().exception(msg)

        # Tries to restore iso without much compromise
        self._set_iso(self.w_conv.f.obj["iso"])
        self.w_conv.f.obj["iso"] = self._get_iso()

        self.w_file.flag_valid = f is not None
        self.changed.emit()

    def _load_lines(self):
        filename = self.w_file.value
        f = self._flines = pyfant.load_kurucz_mol(filename) if os.path.isfile(filename) else None
        self.w_file.flag_valid = f is not None
        self._update_gui_iso()
        return f

    def _update_gui_iso(self):
        """Updates Isotope combobox"""
        f = self._flines
        cb = self.combobox_isotope
        save = cb.blockSignals(True)
        try:
            cb.clear()
            if f is not None:
                if f.__class__  not in (pyfant.FileKuruczMolecule, pyfant.FileKuruczMolecule1):
                    cb.addItem("(all (file is old-format))")
                else:
                    self._isotopes = list(set([line.iso for line in f]))
                    self._isotopes.sort()
                    cb.addItem("(all)")
                    cb.addItems([str(x) for x in self._isotopes])
        finally:
            cb.blockSignals(save)


# class _WPlezPanel(a99.WBase):
#     def __init__(self, parent):
#         a99.WBase.__init__(self, parent.parent_form)
#
#         self.w_conv = parent
#
#         lw = QVBoxLayout()
#         self.setLayout(lw)
#
#         lw.addWidget(self.keep_ref(QLabel("Plez")))


class _WPlezPanel(a99.WBase):
    """
    This panel allows to load a Plez molecular lines file
    """

    @property
    def data(self):
        """Returns FilePlezLineliesBase or None"""
        return self._flines

    @property
    def species(self):
        return self._get_species()

    @species.setter
    def species(self, value):
        self._set_species(value)

    def __init__(self, parent):
        a99.WBase.__init__(self, parent.parent_form)

        self.w_conv = parent

        self._flines = None  # FilePlezLinelistBase
        # list of species names, filled when file is loaded
        self._speciess = []

        lw = QVBoxLayout()
        self.setLayout(lw)

        lw.addWidget(self.keep_ref(QLabel("Plez")))

        lg = QGridLayout()
        lw.addLayout(lg)

        i_row = 0
        a = self.keep_ref(QLabel("Plez molecular lines file"))
        w = self.w_file = a99.WSelectFile(self.parent_form)
        w.changed.connect(self._file_changed)
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Species"))
        w = self.combobox_species = QComboBox()
        w.currentIndexChanged.connect(self.changed)
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        lw.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

    def _set_species(self, value):
        if value is None:
            idx = 0
        else:
            try:
                idx = self._speciess.index(value)+1
            except ValueError:
                idx = 0
        cb = self.combobox_species
        if cb.count() > 0:
            save = cb.blockSignals(True)
            cb.setCurrentIndex(idx)
            cb.blockSignals(save)

    def _get_species(self):
        idx = self.combobox_species.currentIndex()
        if idx <= 0:
            return None
        return self._speciess[idx - 1]

    def _file_changed(self):

        f = None
        try:
            f = self._load_lines()

        except Exception as e:
            self._flines = None
            self._speciess = []
            msg = "Error reading contents of file '{}': '{}'".format(self.w_file.value,
                                                                              a99.str_exc(e))
            self.add_log_error(msg, True)
            a99.get_python_logger().exception(msg)

        # Tries to restore iso without much compromise
        self._set_species(self.w_conv.f.obj["species"])
        self.w_conv.f.obj["species"] = self._get_species()

        self.w_file.flag_valid = f is not None
        self.changed.emit()

    def _load_lines(self):
        filename = self.w_file.value
        f = self._flines = pyfant.load_plez_mol(filename) if os.path.isfile(filename) else None
        self.w_file.flag_valid = f is not None
        self._update_gui_species()
        return f

    def _update_gui_species(self):
        """Updates Species combobox"""
        f = self._flines
        cb = self.combobox_species
        save = cb.blockSignals(True)
        try:
            cb.clear()
            if f is not None:
                # if f.__class__  not in (pyfant.FileKuruczMolecule, pyfant.FileKuruczMolecule1):
                #     cb.addItem("(all (file is old-format))")
                # else:
                self._speciess = list(f.molecules.keys())
                self._speciess.sort()
                cb.addItem("(Please select)")
                cb.addItems([str(x) for x in self._speciess])
        finally:
            cb.blockSignals(save)



class _WConv(a99.WConfigEditor):

    convert_clicked = pyqtSignal()
    open_mol_clicked = pyqtSignal()

    @property
    def flag_hlf(self):
        return self.checkbox_hlf.isChecked()

    @flag_hlf.setter
    def flag_hlf(self, value):
        a99.set_checkbox_value(self.checkbox_hlf, value)

    @property
    def flag_normhlf(self):
        return self.checkbox_normhlf.isChecked()

    @flag_normhlf.setter
    def flag_normhlf(self, value):
        a99.set_checkbox_value(self.checkbox_normhlf, value)

    @property
    def flag_fcf(self):
        return self.checkbox_fcf.isChecked()

    @flag_fcf.setter
    def flag_fcf(self, value):
        a99.set_checkbox_value(self.checkbox_fcf, value)

    @property
    def flag_special_fcf(self):
        return self.checkbox_special_fcf.isChecked()

    @flag_special_fcf.setter
    def flag_special_fcf(self, value):
        a99.set_checkbox_value(self.checkbox_special_fcf, value)

    @property
    def flag_spinl(self):
        return self.checkbox_spinl.isChecked()

    @flag_spinl.setter
    def flag_spinl(self, value):
        a99.set_checkbox_value(self.checkbox_spinl, value)

    @property
    def flag_quiet(self):
        return self.checkbox_quiet.isChecked()

    @flag_quiet.setter
    def flag_quiet(self, value):
        a99.set_checkbox_value(self.checkbox_quiet, value)


    def __init__(self, parent_form):
        a99.WConfigEditor.__init__(self, parent_form)

        self.w_molconsts = parent_form.w_molconsts

        # ## Vertical layout: source and destination stacked
        lsd = self.layout_editor

        # ### Horizontal layout: sources radio buttons, source-specific setup area

        lh = self.keep_ref(QHBoxLayout())
        # a99.set_margin
        lsd.addLayout(lh)

        # #### Vertical layout: source radio group box

        lss = QVBoxLayout()
        lh.addLayout(lss)

        # ##### Source radio buttons
        lss.addWidget(self.keep_ref(QLabel("<b>Source</b>")))
        w = self.w_source = _WSource(self.parent_form)
        w.index_changed.connect(self._source_changed)
        lss.addWidget(w)
        lss.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))


        # #### Adds configuration panels for various sources
        # Only one panel should be visible at a time
        # **Note** The order here doesn't matter
        panels = {}
        panels["HITRAN"] = self.w_hitran = _WHitranPanel(self)
        panels["VALD3"] = self.w_vald3 = _WVald3Panel(self)
        panels["Plez"] = self.w_plez = _WPlezPanel(self)
        panels["Kurucz"] = self.w_kurucz = _WKuruczPanel(self)
        for name in _NAMES:
            ds = _SOURCES[name]
            ds.widget = p = panels[name]
            lh.addWidget(p)
            p.changed.connect(self._panel_changed)

        # ### Flags panel

        fr = self.frame_flags = a99.get_frame()
        lsd.addWidget(fr)
        lfr = self.layout_flags = QVBoxLayout(fr)
        lfr.setSpacing(_LAYSP_V)
        a99.set_margin(lfr, _LAYMN_V)

        la = self.label_flags = QLabel("<b>Flags</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        lfr.addWidget(la)

        lg = QGridLayout()
        lg.setSpacing(_LAYSP_GRID)
        a99.set_margin(lg, _LAYMN_GRID)
        lfr.addLayout(lg)

        i_row = 0

        a = self.keep_ref(QLabel("Calculate 'gf' based on\n"
                                 "Hönl-London factors (HLFs)"))
        w = self.checkbox_hlf = QCheckBox()
        w.setChecked(True)
        w.stateChanged.connect(self.changed)
        w.setToolTip("If selected, will ignore 'loggf' from Kurucz file and\n"
                     "calculate 'gf' using Hönl-London factors formulas taken from Kovacz (1969)")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Apply normalization factor\n"
                                 "for HLFs to add up to 1 (for fixed J)"))
        w = self.checkbox_normhlf = QCheckBox()
        w.setChecked(True)
        w.stateChanged.connect(self.changed)
        w.setToolTip("If selected, calculated 'gf's will be multiplied by\n"
                     "2 / ((2 * J2l + 1) * (2 * S + 1)*(2 - DELTAK))")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Multiply calculated 'gf' by\n"
                                 "Franck-Condon factor"))
        w = self.checkbox_fcf = QCheckBox()
        w.setChecked(True)
        w.stateChanged.connect(self.changed)
        w.setToolTip("If selected, incorporates internally calculated Franck-Condon factor"
                     "into the calculated 'gf'")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Use spin' for branch determination\n"
                                 "(spin'' is always used)"))
        w = self.checkbox_spinl = QCheckBox()
        w.setChecked(True)
        w.stateChanged.connect(self.changed)
        w.setToolTip("If you tick this box, branches P12, P21, Q12, Q21, R21, R12 (i.e., with two numbers) become possible")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Quiet conversion"))
        w = self.checkbox_quiet = QCheckBox()
        w.setChecked(True)
        w.stateChanged.connect(self.changed)
        w.setToolTip("If selected, will show less error messages")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1

        a = self.keep_ref(QLabel("Force 0.1 <= FCF < 1 (debugging option)"))
        w = self.checkbox_special_fcf = QCheckBox()
        w.stateChanged.connect(self.changed)
        w.setToolTip("(*temporary*) If selected, makes 'fcf = fcf * 10 ** -(math.floor(math.log10(fcf)) + 1)'")
        lg.addWidget(a, i_row, 0)
        lg.addWidget(w, i_row, 1)
        i_row += 1


        # ### Output file specification

        w0 = self.w_out = _WSelectSaveFile(self.parent_form)
        w0.wants_autofilename.connect(self._wants_autofilename)
        w0.changed.connect(self._fn_output_changed)
        lsd.addWidget(w0)

        # ### "Convert" button

        lmn = QHBoxLayout()
        lsd.addLayout(lmn)
        lmn.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        b = self.button_convert = QPushButton("&Run Conversion")
        b.clicked.connect(self._convert_clicked)
        lmn.addWidget(b)
        b = self.button_convert = QPushButton("&Open result in mled.py")
        b.clicked.connect(self._open_mol_clicked)
        lmn.addWidget(b)


        # name, property
        self._map = [
            a99.CEMapItem("sourcename", self.w_source),
            a99.CEMapItem("fn_output", self.w_out, propertyname="value"),
            a99.CEMapItem("kurucz_iso", self.w_kurucz, propertyname="iso"),
            a99.CEMapItem("kurucz_fn_input", self.w_kurucz.w_file, propertyname="value"),
            a99.CEMapItem("plez_fn_input", self.w_plez.w_file, propertyname="value"),
            a99.CEMapItem("plez_species", self.w_plez, propertyname="species"),
            a99.CEMapItem("flag_hlf", self),
            a99.CEMapItem("flag_normhlf", self),
            a99.CEMapItem("flag_fcf", self),
            a99.CEMapItem("flag_special_fcf", self),
            a99.CEMapItem("flag_spinl", self),
            a99.CEMapItem("flag_quiet", self),
        ]


    def _do_load(self, fobj):
        a99.WConfigEditor._do_load(self, fobj)
        self.w_kurucz._load_lines()
        # Redundant, but simplest way. After loading data, comboboxes will be positioned
        self._update_gui()

    def _update_gui(self):
        a99.WConfigEditor._update_gui(self)

        idx = self.w_source.index
        for i, ds in enumerate(_SOURCES.values()):
            ds.widget.setVisible(i == idx)

        self._update_w_out_flag_valid()
        # self._source_changed()

    def _update_w_out_flag_valid(self):
        self.w_out.flag_valid = self.w_out.value is not None and len(self.w_out.value) > 0

    def _panel_changed(self):
        """Called when anything in any panel changes"""
        self._update_fobj()
        self.changed.emit()

    def _fn_output_changed(self):
        """Called when anything in any panel changes"""
        self._update_fobj()
        self.changed.emit()
        self._update_w_out_flag_valid()

    def _source_changed(self):
        # if self._flag_updating_gui:
        #     return

            # print("Widget", ds.widget, "is visible?", ds.widget.isVisible())
        self._update_fobj()
        self._update_gui()
        self.changed.emit()

    def _wants_autofilename(self):
        name = _NAMES[self.w_source.index]
        filename = None
        if name == "HITRAN":
            lines = self.w_hitran.data
            if lines:
                filename = "{}.dat".format(lines["header"]["table_name"])

        if filename is None:
            # Default
            filename = a99.new_filename("mol", "dat")
        self.w_out.value = filename
        self._update_fobj()
        self.changed.emit()

    def _convert_clicked(self):
        self.add_log("===BEGIN===")
        try:
            errors = self._validate()

            if len(errors) == 0:
                conv, errors = self._get_conv(errors)

            if len(errors) == 0:
                lines = self._get_lines()
                if lines is None:
                    errors.append("Molecular lines not specified")

            if len(errors) == 0:
                fobj, log = conv.make_file_molecules(lines)

                self._report_conversion(fobj, log)

                if log.flag_ok:
                    fobj.save_as(self.w_out.value)
                    self.add_log("File '{}' generated successfully".format(self.w_out.value))

            else:
                self.add_log_error("Cannot convert:\n  - " + ("\n  - ".join(errors)), True)

        except Exception as e:
            a99.get_python_logger().exception("Conversion failed")
            self.add_log_error("Conversion failed: {}".format(a99.str_exc(e)), True)

        self.add_log("===END===")

    def _report_conversion(self, fobj, log):
        ne = len(log.errors)
        if ne > 0:
            self.add_log("Error messages:")
            self.add_log("\n".join(log.errors))
            self.add_log("{} error message{}".format(ne, "" if ne == 1 else "s"))
        if log.flag_ok:
            if log.num_lines_skipped > 0:
                self.add_log(
                    "Lines filtered out: {}".format(log.num_lines_skipped))
                self.add_log("    Reasons:")
                kv = list(log.skip_reasons.items())
                kv.sort(key=lambda x: x[0])
                for key, value in kv:
                    self.add_log("      - {}: {}".format(key, value))
            ne = log.num_lines - log.num_lines_skipped - fobj.num_lines
            if ne > 0:
                self.add_log(
                    "Lines not converted because of error: {}".format(ne))
            self.add_log(
                "Lines converted: {}/{}".format(fobj.num_lines, log.num_lines))
        else:
            self.add_log_error("Conversion was not possible")

    def _get_conv(self, errors):
        name = self.w_source.source.name

        # (Data source name, Conv class)
        map = dict([("HITRAN", None),
               ("VALD3", None),
               ("Kurucz", pyfant.ConvKurucz),
               ("Plez", pyfant.ConvPlez)])

        cls = None
        conv = None
        try:
            cls = map[name]
        except KeyError:
            pass

        if cls is None:
            errors.append("{}-to-PFANT conversion not implemented yet, sorry".format(name))
        else:
            w = self.w_source.source.widget
            conv = cls(flag_hlf=self.flag_hlf,
                       flag_normhlf=self.flag_normhlf,
                       flag_fcf=self.flag_fcf,
                       flag_special_fcf=self.flag_special_fcf,
                       flag_quiet=self.flag_quiet,
                       )

            if isinstance(conv, pyfant.ConvKurucz):
                conv.iso = w.iso
                conv.flag_spinl = self.flag_spinl
            elif isinstance(conv, pyfant.ConvPlez):
                print("SSSSSSSSSSSSSSSSSSSSSSS", w.species)
                conv.name = w.species

            conv.fcfs = self.w_molconsts.fcfs
            conv.molconsts = self.w_molconsts.f.molconsts

        return conv, errors

    def _get_lines(self):
        """Creates a Conv object and """
        widget_panel = self.w_source.source.widget
        return widget_panel.data

    def _validate(self, ):
        """Validation of GUI. Returns errors"""
        errors = []

        molconsts = self.w_molconsts.f.molconsts

        molconsts_fieldnames_ignore = ["id_molecule", "id_pfantmol", "id_system", "id_statel",
                                       "id_state2l", "pfant_notes", "pfant_name"]
        if molconsts["id_molecule"] is None:
            errors.append("Molecule not selected")

        elif any([value is None for name, value in molconsts.items()
                  if name not in molconsts_fieldnames_ignore]):
            s_none = ", ".join(["'{}'".format(name) for name, value in molconsts.items()
                                if value is None and name not in molconsts_fieldnames_ignore])
            errors.append("There are empty molecular constants: {}".format(s_none))
        if not self.w_out.flag_valid:
            errors.append("Output filename is invalid")



        # Data-source-specific validation

        name = self.w_source.source.name

        if self.flag_fcf and self.w_molconsts.fcfs is None:
            errors.append(
                "Cannot multiply gf's by Franck-Condon Factors, as these are not available in molecular configuration")

        if name == "VALD3":
            if not self.w_vald3.is_molecule:
                errors.append("Need a VALD3 molecule")
        elif name == "Kurucz":
            pass


        return errors

    def _open_mol_clicked(self):
        filename = self.w_out.value
        if len(filename) > 0:
            f = pyfant.FileMolecules()
            f.load(filename)
            vis = pyfant.VisMolecules()
            vis.use(f)

    def _set_checkbox_value(self, w, value):
        save = w.blockSignals(True)
        try:
            w.setChecked(bool(value))
        finally:
            w.blockSignals(save)


class XConvMol(f311.XFileMainWindow):
    def _add_stuff(self):
        # Removed molecular constants database to avoid confusion. Whoever wants it will have to use
        # `moldbed.py` now
        # # Qt stuff tab #0: FileMolDB editor
        # e0 = self.w_moldb = pyfant.WFileMolDB(self)
        # e0.changed.connect(self._on_w_moldb_changed)
        # e0.loaded.connect(self._on_w_moldb_loaded)

        # Qt stuff tab #1: FileMolConsts editor
        e1 = self.w_molconsts = pyfant.WFileMolConsts(self)

        # Qt stuff tab #2: FileConv editor TODO FileConv does not exist yet self.conv a Conv
        e2 = self.w_conv = _WConv(self)

        # self.pages.append(f311.MyPage(text_tab="Molecular constants database",
        #                             cls_save=pyfant.FileMolDB, clss_load=(pyfant.FileMolDB,), wild="*.sqlite", editor=e0, flag_autosave=True))

        self.pages.append(f311.MyPage(text_tab="Molecular constants",
                                    cls_save=pyfant.FileMolConsts, clss_load=(pyfant.FileMolConsts,), wild="*.py", editor=e1))

        self.pages.append(f311.MyPage(text_tab="Conversion",
                                    cls_save=pyfant.FileConfigConvMol, clss_load=(pyfant.FileConfigConvMol,), wild="*.py", editor=e2))

        self.setWindowTitle("(to) PFANT Molecular Lines Converter")
        self.installEventFilter(self)

    def eventFilter(self, obj_focused, event):
        if event.type() == QEvent.KeyPress:
            # To help my debugging: Configures for Kurucz OH conversion
            if event.key() == Qt.Key_F12:
                self.w_moldb.combobox_select_molecule.setCurrentIndex(7)
                self.w_moldb.combobox_select_pfantmol.setCurrentIndex(1)
                self.w_moldb.combobox_select_state.setCurrentIndex(6)
                self.w_moldb.combobox_select_system.setCurrentIndex(0)
                self.w_moldb.None_to_zero()
                self.w_source._buttons[2].setChecked(True)
                self.source_changed()
                self.w_kurucz.w_file.edit.setText("ohaxupdate.asc")
                self.tabWidget.setCurrentIndex(1)
        return False

    # TODO cleanup moldb part once I am sure it is no longer needed def _on_w_moldb_changed(self):
    #     self.w_molconsts.set_moldb(self.w_moldb.f)

    def _on_w_moldb_loaded(self):
        self.w_molconsts.set_moldb(self.w_moldb.f)

    def _on_w_molconsts_changed(self):
        pass

    def _on_w_conv_changed(self):
        pass



# # Maybe I can use this one day. But the purpose was to validate elements, but confusing CO with Co is not nice

# import f311.ftpyfant as pf
# import a99
#
#
# __all__ = ["is_element"]
#
#
# __SYMBOLS_ADJUSTED = None
#
#
# def SYMBOLS_ADJUSTED():
#     global __SYMBOLS_ADJUSTED
#     if __SYMBOLS_ADJUSTED is None:
#         __SYMBOLS_ADJUSTED = [pyfant.adjust_atomic_symbol(x) for x in a99.symbols]
#     return __SYMBOLS_ADJUSTED
#
# def is_element(symbol):
#     """Returns True if symbol is an atomic element according to convention in adjust_atomic_symbol()
#
#     >>> is_element("CO")
#     False
#     >>> is_element("       fe   ")
#     True
#     """
#     return pyfant.adjust_atomic_symbol(symbol) in SYMBOLS_ADJUSTED()
#



#!/usr/bin/env python

"""Conversion of molecular lines data to PFANT format"""

import sys
import argparse
import a99
import logging
import os
import pyfant


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":

    deffn = pyfant.FileMolDB.default_filename

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('--fn_molconsts', type=str, help='File name for {}'.format(pyfant.FileMolConsts.description),
                        default=pyfant.FileMolConsts.default_filename, nargs='?')
    parser.add_argument('--fn_config', type=str, help='File name for {}'.format(pyfant.FileConfigConvMol.description),
                        default=pyfant.FileConfigConvMol.default_filename, nargs='?')
    args = parser.parse_args()

    _fnfn = [args.fn_molconsts, args.fn_config]
    fnfn = [None if fn is None or not os.path.isfile(fn) else fn for fn in _fnfn]

    app = a99.get_QApplication([])
    form = pyfant.XConvMol()
    form.load_many(fnfn)
    form.show()
    sys.exit(app.exec_())





class ParseError(Exception):
  pass

class FileConsistencyError(Exception):
  """Things that must match within file don't."""
  pass

class FailedError(Exception):
  """Execution failed for one of the executables"""


#!/usr/bin/env python

"""
Retrieves molecular lines from the HITRAN database [Gordon2016]

This script uses web scraping and the HAPI to save locally molecular lines from the HITRAN database.

While the HAPI provides the downloading facility, web scraping is used to get the lists of molecules
and isotopologues from the HITRAN webpages and get the IDs required to run the HAPI query.

The script is typically invoked several times, each time with an additional argument.

References:

[Gordon2016] I.E. Gordon, L.S. Rothman, C. Hill, R.V. Kochanov, Y. Tan, P.F. Bernath, M. Birk,
    V. Boudon, A. Campargue, K.V. Chance, B.J. Drouin, J.-M. Flaud, R.R. Gamache, J.T. Hodges,
    D. Jacquemart, V.I. Perevalov, A. Perrin, K.P. Shine, M.-A.H. Smith, J. Tennyson, G.C. Toon,
    H. Tran, V.G. Tyuterev, A. Barbe, A.G. Császár, V.M. Devi, T. Furtenbacher, J.J. Harrison,
    J.-M. Hartmann, A. Jolly, T.J. Johnson, T. Karman, I. Kleiner, A.A. Kyuberis, J. Loos,
    O.M. Lyulin, S.T. Massie, S.N. Mikhailenko, N. Moazzen-Ahmadi, H.S.P. Müller, O.V. Naumenko,
    A.V. Nikitin, O.L. Polyansky, M. Rey, M. Rotger, S.W. Sharpe, K. Sung, E. Starikova,
    S.A. Tashkun, J. Vander Auwera, G. Wagner, J. Wilzewski, P. Wcisło, S. Yu, E.J. Zak,
    The HITRAN2016 Molecular Spectroscopic Database, J. Quant. Spectrosc. Radiat. Transf. (2017).
    doi:10.1016/j.jqsrt.2017.06.038.
"""

import argparse
import logging
import a99
import tabulate
import sys
import os
from f311 import hapi
import airvacuumvald as avv
import pyfant


_DEF_T = '(molecular formula)'
_DEF_M = '(lists molecules)'
_DEF_I = '(lists isotopologues)'
_DEF_LL = None


if __name__ == "__main__":
    a99.logging_level = logging.INFO
    a99.flag_log_file = True
    script_name = os.path.basename(__file__)

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('-t', type=str, help='Table Name',
                        default=_DEF_T, nargs=1)
    parser.add_argument('M', type=str, help='HITRAN molecule number',
                        default=_DEF_M, nargs='?')
    parser.add_argument('I', type=str,
                        help='HITRAN isotopologue number (not unique, starts over at each molecule)',
                        default=_DEF_I, nargs='?')
    parser.add_argument('llzero', type=float,
                        help='Initial wavelength (Angstrom)',
                        default=_DEF_LL, nargs='?')
    parser.add_argument('llfin', type=float,
                        help='Final wavelength (Angstrom)',
                        default=_DEF_LL, nargs='?')
    # parser.add_argument('llzero', type=str,
    # help='initial wavelength (',
    # default='()', nargs='?')
    # parser.add_argument('I', type=str,
    # help='HITRAN isotopologue number (not unique, starts over at each molecule',
    # default='()', nargs='?')
    args = parser.parse_args()

    db = pyfant.FileHitranDB()
    db.init_default()
    if args.M == _DEF_M:
        print()
        print("\n".join(a99.format_h1("List of all HITRAN molecules")))
        print()
        # Lists all molecules from local cache of HITRAN molecules
        print(tabulate.tabulate(*a99.cursor_to_data_header(db.query_molecule())))
        print()
        print("Now, to list isotopologues for a given molecule, please type:")
        print()
        print("    {} <molecule ID>".format(script_name))
        print()
        print("where <molecule ID> is one of the IDs listed above.")
    elif args.I == _DEF_I:
        row0 = db.query_molecule(ID=args.M).fetchone()
        if row0 is None:
            a99.print_error("Molecule ID {} not found".format(args.M))
            sys.exit()

        print()
        print("\n".join(a99.format_h1("List of all isotopologues for molecule '{Formula}' ({Name})".
                                     format(**row0))))
        print()
        # Lists all molecules from local cache of HITRAN molecules
        print(tabulate.tabulate(*a99.cursor_to_data_header(
            db.query_isotopologue(**{"molecule.ID": args.M}))))
        print()
        print("Now, to download lines, please type:")
        print()
        print("    {} {} <isotopologue ID> <llzero> <llfin>".format(script_name, args.M))
        print()
        print("where <isotopologue ID> is one the numbers from the 'ID' column above,")
        print()
        print("and [<llzero>, <llfin>] defines the wavelength interval in Angstrom.")
        print()
    else:
        row0 = db.query_molecule(ID=args.M).fetchone()
        if row0 is None:
            a99.print_error("Molecule ID {} not found".format(args.M))
        else:
            kwargs = {"molecule.id": args.M, "isotopologue.ID": args.I}
            row1 = db.query_isotopologue(**kwargs).fetchone()
            if row1 is None:
                a99.print_error("Isotopologue ({}, {}) not found".format(args.M, args.I))
                sys.exit()

            print()
            print("\n".join(a99.format_h1("Isotopologue selected:")))
            print()
            print(tabulate.tabulate(list(row1.items()), ["Field name", "Value"]))
            print()

            if args.llzero == _DEF_LL or args.llfin == _DEF_LL:
                print("To download lines for this isotopologue, please type:")
                print()
                print("    {} {} {} <llzero> <llfin>".format(script_name, args.M, args.I))
                print()
                print("where [<llzero>, <llfin>] makes the wavelength interval in Angstrom.")
                print()
                sys.exit()

            print("Wavelength interval (air): [{llzero}, {llfin}] Angstrom".format(**args.__dict__))
            if args.llzero >= args.llfin:
                a99.print_error("Initial wavelength must be lower than Final wavelength")
                sys.exit()

            # Here will fetch data
            wn0 = avv.air_to_vacuum(1e8/args.llzero)
            wn1 = avv.air_to_vacuum(1e8/args.llfin)
            print("Wavenumber interval (vacuum): [{}, {}] cm**-1".format(wn1, wn0))

            table_name = args.t if args.t != _DEF_T else row1["Formula"]
            print("Table name: '{}'".format(table_name))
            print()
            print("Fetching data...")
            print("===")
            print("=== BEGIN messages from HITRAN API ===")
            print("===")
            hapi.fetch(table_name, int(args.M), int(args.I), wn1, wn0)

            print("===")
            print("=== END messages from HITRAN API ===")
            print("===")
            print("...done")
            print("Please check files '{0}.header', '{0}.data'".format(table_name))





    # kwargs = {}
    # if not args.ID == "(all)":
    #     kwargs["molecule.ID"] = args.ID
    #
    # pyfant.hitrandb.print_isotopologues(**kwargs)


from f311 import FilePy
import a99
from ..molconsts import MolConsts
from . import FileMolDB

__all__ = ["FileMolConsts"]

@a99.froze_it
class FileMolConsts(FilePy):
    """Molecular constants config file (Python code)"""

    default_filename = "configmolconsts.py"
    attrs = ["fn_moldb", "molconsts"]
    editors = ["mced.py"]

    def __init__(self):
        FilePy.__init__(self)
        self.fn_moldb = FileMolDB.default_filename
        self.molconsts = MolConsts()

    def _do_load(self, filename):
        module = a99.import_module(filename)
        try:
            self.fn_moldb = module.fn_moldb
        except AttributeError:
            pass
        self._copy_attr(module, "molconsts", MolConsts)

    def _do_save_as(self, filename):
        with open(filename, "w") as h:
            h.write("{}\n"
                    "from pyfant import MolConsts\n"
                    "\n"
                    "fn_moldb = '{}'\n"
                    "\n"
                    "molconsts = {}\n".format(self._get_header(), self.fn_moldb, self.molconsts))

    def init_default(self):
        pass


from f311 import FileSpectrum, Spectrum
import numpy as np

class FileSpectrumPfant(FileSpectrum):
    """
    PFANT Spectrum (`pfant` output)

    This file alternates a "header" line where most of the information is
    repeated, and a "values" line, with the values of the flux
    corresponding to the lambda interval lzero-lfin
    """

    default_filename = "flux.norm"

    @property
    def l0(self):
        if self.spectrum is not None:
            if self.spectrum.x is not None:
                return self.spectrum.x[0]
        return None

    @property
    def lf(self):
        if self.spectrum is not None:
            if self.spectrum.x is not None:
                return self.spectrum.x[-1]
        return None

    @property
    def pas(self):
        if self.spectrum is not None:
            return self.spectrum.delta_lambda
        return None

    def __init__(self):
        FileSpectrum.__init__(self)

        self.tit = ""
        self.tetaef = 0.
        self.glog = 0.
        self.asalog = 0.
        self.modeles_nhe = 0.
        self.amg = 0.
        self.echx = 0.
        self.echy = 0.
        self.fwhm = 0.

    def _do_load(self, filename):
#    55                            1.17895        1.55000       -0.54000        0.08511        0.00000    3083.0    3630.0    3083.0    3093.0      2001        0.00500        5.00000        1.00000        0.12000
# 12345                    123456789012345               123456789012345               123456789012345          1234567890          1234567890          123456789012345               123456789012345
#      12345678901234567890               123456789012345               123456789012345               1234567890          1234567890          1234567890               123456789012345               123456789012345
# ikeytot                                                                                             l0                                                pas
# 1234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
# 0         10        20        30        40        50        60        70        80        90        100       110       120       130       140       150

        with open(filename, 'r') as h:
            # f_header = ff.FortranRecordReader(
            #     "(i5, a20, 5f15.5, 4f10.1, i10, 4f15.5)")
            i = 0
            y = []
            sp = self.spectrum = Spectrum()

            while True:
                s = h.readline()
                if i == 0:
                    ikeytot = int(s[0:5])
                    l0 = float(s[100:110])
                    pas = float(s[150:165])

                    # vars_ = f_header.read(s)  # This is actually quite slow, gotta avoid it
                    # [ikeytot,
                    #  _,
                    #  _,
                    #  _,
                    #  _,
                    #  _,
                    #  _,
                    #  l0,
                    #  _,
                    #  _,  # lzero,
                    #  _,  # lfin,
                    #  _,  # itot
                    #  pas,
                    #  _,
                    #  _,
                    #  _] = vars_
                    #
                    # # [self.ikeytot,
                    # #  self.tit,
                    # #  self.tetaef,
                    # #  self.glog,
                    # #  self.asalog,
                    # #  self.modeles_nhe,
                    # #  self.amg,
                    # #  self.l0,
                    # #  self.lf,
                    # #  _,  # lzero,
                    # #  _,  # lfin,
                    # #  _,  # itot
                    # #  self.pas,
                    # #  self.echx,
                    # #  self.echy,
                    # #  self.fwhm] = vars_

                #itot = vars[11]

                # if False:
                # # Will have to compile a different formatter depending on itot
                #     if i == 0 or itot != last_itot:
                #       f_flux = ff.FortranRecordReader("(%df15.5)" % itot)
                #
                #     v = f_flux.read(h.readline())
                # else:
                # Will see if can read faster than Fortran formatter
                s = h.readline()

                # Note: last character of s is "\n"
                v = [float(s[0 + j:15 + j]) for j in range(0, len(s) - 1, 15)]

                # print len(v)
                # print "==========================="
                # print v
                # print i, self.ikeytot

                # if i < self.ikeytot - 1:
                if i < ikeytot - 1:
                    # Last point is discarded because pfant writes reduntantly:
                    # last point of iteration ikey is the same as first point of
                    # iteration ikey+1
                    # y = y + v  # update: taking all points calculated yes [:-1]
                    y = y + v[:-1]
                else:
                    # ...except for in the last calculation interval
                    # (then the last point is used).
                    y = y + v
                    break

                i += 1

                #last_itot = itot

        # Lambdas

        # sp.x = np.array([self.l0 + k * self.pas for k in range(0, len(y))])

        sp.x = np.array([l0 + k * pas for k in range(0, len(y))])

        sp.y = np.array(y)

#        logging.debug("Just read PFANT Spectrum '%s'" % filename)

    def _do_save_as(self, filename):
#    55                            1.17895        1.55000       -0.54000        0.08511        0.00000    3083.0    3630.0    3083.0    3093.0      2001        0.00500        5.00000        1.00000        0.12000
# 12345                    123456789012345               123456789012345               123456789012345          1234567890          1234567890          123456789012345               123456789012345
#      12345678901234567890               123456789012345               123456789012345               1234567890          1234567890          1234567890               123456789012345               123456789012345
# ikeytot                                                                                             l0                                                pas
# 1234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
# 0         10        20        30        40        50        60        70        80        90        100       110       120       130       140       150

        with open(filename, 'w') as h:


            fmt = "{:5d}{:20s}" \
                  "{:15.5f}{:15.5f}{:15.5f}{:15.5f}{:15.5f}" \
                  "{:10.1f}{:10.1f}{:10.1f}{:10.1f}" \
                  "{:10d}" \
                  "{:15.5f}{:15.5f}{:15.5f}{:15.5f}\n"

            ikeytot = 1
            itot = len(self.spectrum.y)

            h.write(fmt.format(
                ikeytot,
                self.tit,
                self.tetaef,
                self.glog,
                self.asalog,
                self.modeles_nhe,
                self.amg,
                self.l0,
                self.lf,
                self.l0,
                self.lf,
                itot,
                self.pas,
                self.echx,
                self.echy,
                self.fwhm
            ))

            h.write("".join(["{:15.5f}".format(a) for a in self.spectrum.y])+"\n")


class FileSpectrumNulbad(FileSpectrum):
    """
    PFANT Spectrum (`nulbad` output)

    This file is a two-column text file with two lines of comment at the beginning
    """

    def _do_load(self, filename):
        x = []
        y = []
        with open(filename, 'r') as h:
            sp = self.spectrum = Spectrum()
            # -- row 01 --
            # Original format: ('#',A,'Tef=',F6.3,X,'log g=',F4.1,X,'[M/H]=',F5.2,X,F5.2)
            # s_header0 = struct.Struct("1x 20s 4x 6s 7x 4s 7x 5s 1x 5s")
            s = h.readline()

            if not s.startswith("#"):
                raise RuntimeError("Not a nulbad output file")

            # [sp.tit, sp.tetaef, sp.glog, sp.asalog, sp.amg] = \
            #     s_header0.unpack_from(s)
            # [sp.tetaef, sp.glog, sp.asalog, sp.amg] = \
            #     list(map(float, [sp.tetaef, sp.glog, sp.asalog, sp.amg]))

            # -- row 02 --
            # Original format: ('#',I6,2X,'0. 0. 1. 1. Lzero =',F10.2,2x,'Lfin =', &
            # F10.2,2X,'PAS =',F5.2,2x,'FWHM =',F5.2)
            # s_header1 = struct.Struct("1x 6s 21x 10s 8x 10s 7x 5s 8x 5s")
            s = h.readline()

            if not s.startswith("#"):
                raise RuntimeError("Not a nulbad output file")

            # [_, sp.l0, sp.lf, sp.pas, sp.fwhm] = \
            #     s_header1.unpack_from(s)
            # [sp.l0, sp.lf, sp.pas, sp.fwhm] = \
            #     list(map(float, [sp.l0, sp.lf, sp.pas, sp.fwhm]))
            #n = int(n)


            # -- rows 03 ... --
            #
            # Examples:
            #    4790.0000000000000       0.99463000000000001
            #    4790.0400000000000        2.0321771294932130E-004

            # pattern = re.compile(r"^\s+([0-9.E+-]+)\s+([0-9.E+-]+)")
            while True:
                s = h.readline().strip()
                if not s:
                  break

                a, b = [float(z) for z in s.split()]
                # if match is None:
                #     raise ParseError('Row %d of file %s is invalid' % (i + 3, filename))
                # a, b = map(float, match.groups())

                x.append(a)
                y.append(b)

        sp.x = np.array(x)
        sp.y = np.array(y)

#        logging.debug("Just read NULBAD Spectrum '%s'" % filename)




#!/usr/bin/env python
"""
Runs the four Fortran binaries in sequence: `innewmarcs`, `hydro2`, `pfant`, `nulbad`

Check session directory "session-<number>" for log files.
"""

import argparse
import pyfant
import a99
import logging


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )

    names = pyfant.Conf().opt.get_names() # option names

    for name in names:
        # name = name.replace('_', '-')
        parser.add_argument("--"+name, type=str, help='')

    args = parser.parse_args()


    # Configuration for Python logging messages.
    logger = a99.get_python_logger()

    c = pyfant.Combo()
    c.conf.flag_log_file = True  # Configuration for Fortran messages
    c.conf.flag_log_console = True  # "
    c.conf.flag_output_to_dir = False  # Will generate outputs in current directory

    for name in names:
        x = args.__getattribute__(name)
        if x is not None:
            c.conf.opt.__setattr__(name, x)

    c.run()
    logger.info("Session directory: %s" % c.conf.sid.dir)



# -*- FilePy: FileAbXFwhm -*-
#
# Specification of differential abundances for each chemical.
# Differential abundance is a number to add to original value in abonds.dat.
# All lists of abundances for each element must have the same length.
ab = {"Ca": [-.3, 0, .3, .5],
      "Si": [-.3, 0, .3, .5]}

# Name for each pfant run. They will be part of spectrum file names.
# This is optional. If not used, sequential numbers will be used instead.
# Example: pfant_names = ["A", "B", "C", "D"]
pfant_names = []

# Convolutions specification for fwhm parameter:
# [first value, last value, step]
conv = [0.08, 0.6,  0.04]



__all__ = ["FileAbsoru2"]

from f311 import DataFile


class FileAbsoru2(DataFile):
    """PFANT "Absoru2" file

    Reader/writes not implemented (will be implemented when there is the need for so)"""

    default_filename = "absoru2.dat"

    def __init__(self):
        DataFile.__init__(self)

    def _do_load(self, filename):
        raise NotImplementedError("This class is a stub ATM")

    def _do_save_as(self, filename):
        raise NotImplementedError("This class is a stub ATM")
